/*
 * Copyright (c) 2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/error.h>

#include "csv.h"

/*
 * RFC 4180 CSV parser
 */

#define EMPTY_FIELD(next_state) \
	do { \
		ASSERT(!quoted); \
		if (field == nptrs) \
			return -ENOMEM; \
		ptrs[field] = &input[i]; \
		sizes[field] = 0; \
		needs_unquote[field] = false; \
		field++; \
		state = (next_state); \
	} while (0)

#define START_FIELD(idx, is_quoted) \
	do { \
		ASSERT(!quoted); \
		if (field == nptrs) \
			return -ENOMEM; \
		quoted = (is_quoted); \
		ptrs[field] = &input[idx]; \
		sizes[field] = 0; \
		needs_unquote[field] = false; \
		state = CSV_MOF; \
	} while (0)

#define END_FIELD(idx, next_state, quoted_expected) \
	do { \
		ASSERT3U(quoted, ==, quoted_expected); \
		sizes[field] = ((uintptr_t) &input[idx]) - ((uintptr_t) ptrs[field]); \
		state = (next_state); \
		quoted = false; \
		field++; \
	} while (0)

enum csv_state {
	CSV_BOF, /* begining of field */
	CSV_MOF, /* middle of field */
	CSV_MOF_QUOTE, /* middle of field, looking for second quote */
	CSV_NL, /* expecting a \n */
	CSV_DONE,
};

ssize_t csv_parse_line(const char *input, size_t len, size_t *nfields,
		       const char **ptrs, size_t *sizes, bool *needs_unquote,
		       size_t nptrs)
{
	enum csv_state state = CSV_BOF;
	bool quoted;
	size_t field;
	size_t i;

	if (!len)
		return -EEOF;

	quoted = false;
	field = 0;

	for (i = 0; (i < len) && (state != CSV_DONE); i++) {
		char c = input[i];

		switch (c) {
			case '"':
				switch (state) {
					case CSV_BOF:
						/* start of a quoted field */
						START_FIELD(i + 1, true);
						break;
					case CSV_MOF:
						if (!quoted)
							return -EINVAL;
						state = CSV_MOF_QUOTE;
						break;
					case CSV_MOF_QUOTE:
						/* second quote in a row */
						ASSERT(quoted);
						*needs_unquote = true;
						state = CSV_MOF;
						break;
					case CSV_NL:
						return -EINVAL;
					case CSV_DONE:
						break; /* unreachable */
				}
				break;
			case ',':
				switch (state) {
					case CSV_BOF:
						/* empty field */
						EMPTY_FIELD(CSV_BOF);
						break;
					case CSV_MOF:
						if (!quoted)
							END_FIELD(i, CSV_BOF, false);
						break;
					case CSV_MOF_QUOTE:
						/* end of field */
						END_FIELD(i - 1, CSV_BOF, true);
						break;
					case CSV_NL:
						return -EINVAL;
					case CSV_DONE:
						break; /* unreachable */
				}
				break;
			case '\r':
				switch (state) {
					case CSV_BOF:
						/* empty field @ end of line */
						EMPTY_FIELD(CSV_NL);
						break;
					case CSV_MOF:
						/* end of unquoted field @ EOL */
						if (!quoted)
							END_FIELD(i, CSV_NL, false);
						break;
					case CSV_MOF_QUOTE:
						/* end of quoted field @ EOL */
						END_FIELD(i - 1, CSV_NL, true);
						break;
					case CSV_NL:
						return -EINVAL;
					case CSV_DONE:
						break; /* unreachable */
				}
				break;
			case '\n':
				switch (state) {
					case CSV_BOF:
						/* empty field @ end of line */
						EMPTY_FIELD(CSV_DONE);
						break;
					case CSV_MOF:
						/* end of unquoted field @ EOL */
						if (!quoted)
							END_FIELD(i, CSV_DONE, false);
						break;
					case CSV_MOF_QUOTE:
						/* end of quoted field @ EOL */
						END_FIELD(i - 1, CSV_DONE, true);
						break;
					case CSV_NL:
						state = CSV_DONE;
						break;
					case CSV_DONE:
						break; /* unreachable */
				}
				break;
			default:
				switch (state) {
					case CSV_BOF:
						/* start of unquoted field */
						START_FIELD(i, false);
						break;
					case CSV_MOF:
						break;
					case CSV_MOF_QUOTE:
					case CSV_NL:
						return -EINVAL;
					case CSV_DONE:
						break; /* unreachable */
				}
				break;
		}
	}

	switch (state) {
		case CSV_BOF:
			EMPTY_FIELD(CSV_DONE);
			break;
		case CSV_MOF:
			END_FIELD(i, CSV_DONE, false);
			break;
		case CSV_MOF_QUOTE:
			END_FIELD(i - 1, CSV_DONE, true);
			break;
		case CSV_NL:
			return -EINVAL; /* ended with \r only */
		case CSV_DONE:
			break;
	}

	*nfields = field;

	return i;
}

