/*
 * Copyright (c) 2020-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/types.h>
#include <jeffpc/error.h>

#include <ctype.h>
#include <string.h>

#include <hlog/maidenhead.h>

/*
 * Get the lower or upper bound for any char in a grid square.  The entire
 * string is divided into longitude/latitude character pairs.  Each grouping
 * is either numeric (0..9) or alphabetic (A..R or A..X).  The odd groupings
 * are always numeric, while the even groupings are always alphabetic.  The
 * first grouping is not only alphabetic but has a special range (A..R
 * instead of A..X used by other alphabetic groupings).
 *
 * For example,
 *
 * input:  FN42AA00BB11...
 * group:  001122334455...
 * min:    AA00AA00AA00...
 * max:    RR99XX99XX99...
 */
static inline char group2bound(size_t grp, bool min)
{
	bool numeric = (grp & 1);

	if (min)
		return numeric ? '0' : 'A';
	else if (!grp)
		return 'R';
	else
		return numeric ? '9' : 'X';
}

static bool encode_grid(char *letter, double *last_step, double *val,
			char minc, char maxc)
{
	const size_t nsteps = maxc - minc + 1;
	const double step = *last_step / nsteps;
	const unsigned int whole = *val / step;

	ASSERT3U(whole, <, nsteps);

	*letter = minc + whole;
	*last_step = step;
	*val -= whole * step;

	return true;
}

bool maidenhead_encode(char *grid, size_t gridsize, double lat, double lon)
{
	double lat_step, lon_step;
	size_t i;

	if (!grid || (gridsize < 3))
		return false; /* need at least 2 chars + nul */
	if ((lat < -90) || (lat > +90))
		return false;
	if ((lon < -180) || (lon > +180))
		return false;

	lat += 90.;
	lon += 180.;
	lat_step = 180.;
	lon_step = 360.;

	for (i = 0; i < (gridsize - 1) / 2; i++) {
		const char minc = group2bound(i, true);
		const char maxc = group2bound(i, false);

		if (!encode_grid(&grid[2 * i + 0], &lon_step, &lon, minc, maxc))
			return false;
		if (!encode_grid(&grid[2 * i + 1], &lat_step, &lat, minc, maxc))
			return false;
	}

	grid[2 * i] = '\0';

	return true;
}

static inline bool decode_grid(double *loc, double *last_step, char c,
			       char minc, char maxc)
{
	/* uppercase input to match expected range */
	c = toupper(c);

	if ((c < minc) || (c > maxc))
		return false;

	*last_step /= maxc - minc + 1;
	*loc += (c - minc) * *last_step;

	return true;
}

bool maidenhead_decode(const char *grid, double *lat_r, double *lon_r)
{
	double lat, lon;
	double lat_step, lon_step;
	size_t len;
	size_t i;

	if (!grid)
		return false;

	len = strlen(grid);

	/* at least 2 chars and multiple of 2 chars */
	if ((len < 2) || ((len % 2) != 0))
		return false;

	lat = -90.;
	lon = -180.;
	lat_step = 180.;
	lon_step = 360.;

	for (i = 0; i < len; i += 2) {
		const char minc = group2bound(i / 2, true);
		const char maxc = group2bound(i / 2, false);

		if (!decode_grid(&lon, &lon_step, grid[i + 0], minc, maxc))
			return false;
		if (!decode_grid(&lat, &lat_step, grid[i + 1], minc, maxc))
			return false;
	}

	*lat_r = lat + lat_step / 2;
	*lon_r = lon + lon_step / 2;

	return true;
}
