/*
 * Copyright (c) 2020-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/buffer.h>
#include <jeffpc/io.h>
#include <jeffpc/rand.h>
#include <jeffpc/mem.h>

#include "qso_impl.h"
#include <hlog/bands.h>
#include <hlog/util.h>

/*
 * FIXME: ELF files trivially handle weak symbols but for whatever reason
 * #pragma weak doesn't work on macOS and/or MachO
 */
//#define HAVE_PRAGMA_WEAK

struct qso *qso_alloc(void)
{
	struct qso *qso;

	qso = zalloc(sizeof(struct qso));
	if (!qso)
		return NULL;

	refcnt_init(&qso->refcnt, 1);

	qso_reset(qso);

	return qso;
}

static void qso_side_free(struct qso_side *side)
{
	val_putref(side->station_call);
	val_putref(side->operator_call);
	val_putref(side->power);
	val_putref(side->freq);
	val_putref(side->mode);
	val_putref(side->rst);
	val_putref(side->name);
	val_putref(side->qth);
	val_putref(side->grid);
	val_putref(side->country);
	val_putref(side->state);
	val_putref(side->county);
	val_putref(side->continent);
	val_putref(side->dxcc);
	val_putref(side->itu);
	val_putref(side->cqz);
	val_putref(side->seq);
}

void qso_free(struct qso *qso)
{
	if (!qso)
		return;

	val_putref(qso->comment);

	qso_side_free(&qso->tx);
	qso_side_free(&qso->rx);

	nvl_putref(qso->additional);

	free(qso);
}

int qso_save(const struct qso *qso, int data_dir)
{
	char tmpname[FILENAME_MAX];
	char qsoname[FILENAME_MAX];
	struct buffer *buf;
	int ret;

	ASSERT(!xuuid_is_null(&qso->uuid));

	buf = qso_pack_hlq(qso);
	if (IS_ERR(buf))
		return PTR_ERR(buf);

	snprintf(tmpname, sizeof(tmpname), "tmp/%s.hlq", qso->uuid_str);
	snprintf(qsoname, sizeof(qsoname), "qso/%s.hlq", qso->uuid_str);

	ret = xunlinkat(data_dir, tmpname, 0);
	if (ret && (ret != -ENOENT))
		goto err;

	ret = write_file_at(data_dir, tmpname, buffer_data(buf),
			    buffer_size(buf));
	if (ret)
		goto err;

	/* TODO: sync the file */

	ret = xrenameat(data_dir, tmpname, data_dir, qsoname);

	/* TODO: fsync the dir */

err:
	buffer_free(buf);

	return ret;
}

static void qso_side_reset(struct qso_side *side)
{
	qso_clear_station_call(side);
	qso_clear_operator_call(side);
	qso_clear_power(side);
	qso_clear_freq(side);
	qso_clear_mode(side);
	qso_clear_rst(side);
	qso_clear_name(side);
	qso_clear_qth(side);
	qso_clear_grid(side);
	qso_clear_country(side);
	qso_clear_state(side);
	qso_clear_county(side);
	qso_clear_continent(side);
	qso_clear_zone_dxcc(side);
	qso_clear_zone_itu(side);
	qso_clear_zone_cqz(side);
	qso_clear_seq(side);
}

void qso_reset(struct qso *qso)
{
	qso_clear_datetime(qso, true);
	qso_clear_datetime(qso, false);
	qso_clear_comment(qso);

	qso_side_reset(&qso->tx);
	qso_side_reset(&qso->rx);

	nvl_putref(qso->additional);
	qso->additional = nvl_alloc();
	ASSERT(!IS_ERR(qso->additional));

	xuuid_generate(&qso->uuid);
	xuuid_unparse(&qso->uuid, qso->uuid_str);

	qso->tx.qso = qso;
	qso->rx.qso = qso;
}

static void qso_copy_side(struct qso_side *dst, struct qso_side *src)
{
	dst->station_call = val_getref(src->station_call);
	dst->operator_call = val_getref(src->operator_call);
	dst->power = val_getref(src->power);
	dst->freq = val_getref(src->freq);
	dst->mode = val_getref(src->mode);
	dst->rst = val_getref(src->rst);
	dst->name = val_getref(src->name);
	dst->qth = val_getref(src->qth);
	dst->grid = val_getref(src->grid);
	dst->country = val_getref(src->country);
	dst->state = val_getref(src->state);
	dst->county = val_getref(src->county);
	dst->continent = val_getref(src->continent);
	dst->dxcc = val_getref(src->dxcc);
	dst->itu = val_getref(src->itu);
	dst->cqz = val_getref(src->cqz);
	dst->seq = val_getref(src->seq);
}

void qso_copy(struct qso *dst, struct qso *src)
{
	qso_reset(dst);

	dst->uuid = src->uuid;
	memcpy(dst->uuid_str, src->uuid_str, sizeof(src->uuid_str));
	dst->start = src->start;
	dst->end = src->end;
	dst->comment = val_getref(src->comment);

	qso_copy_side(&dst->tx, &src->tx);
	qso_copy_side(&dst->rx, &src->rx);

	nvl_putref(dst->additional);
	dst->additional = nvl_alloc();
	ASSERT(!IS_ERR(dst->additional));
	nvl_merge(dst->additional, src->additional);
}

/*
 * QSO functions
 */

/* start & end dates */
/*
 * Try to parse using different date & time formats.
 *
 * We make up to two passes through the list of formats.  If the first pass
 * resulted in successful parsing, we advance the input past the parsed
 * values and try again.  The goal is to allow parsing of input with both
 * date and time (e.g., "2020-07-10 02:43:28").
 */
static int __qso_set_datetime(struct time *time, char *str)
{
	static struct {
		const char *fmt;
		enum time_parts have;
	} formats[] = {
		/* ISO 8601 */
		{ "%Y-%m-%dT%H:%MZ",    TIME_PART_DATE | TIME_PART_TIME_HM },
		{ "%Y-%m-%dT%H:%M:%SZ", TIME_PART_DATE | TIME_PART_TIME_HMS },
		{ "%Y%m%dT%H%MZ",       TIME_PART_DATE | TIME_PART_TIME_HM },
		{ "%Y%m%dT%H%M%SZ",     TIME_PART_DATE | TIME_PART_TIME_HMS },

		/* human readable */
		{ "%Y-%m-%d",	TIME_PART_DATE },
		{ "%Y%m%d",	TIME_PART_DATE },
		{ "%H:%M:%SZ",	TIME_PART_TIME_HMS },
		{ "%H:%M:%S",	TIME_PART_TIME_HMS },
		{ "%H:%MZ",	TIME_PART_TIME_HM },
		{ "%H:%M",	TIME_PART_TIME_HM },
		{ "%H%M%SZ",	TIME_PART_TIME_HMS },
		{ "%H%M%S",	TIME_PART_TIME_HMS },
		{ "%H%MZ",	TIME_PART_TIME_HM },
		{ "%H%M",	TIME_PART_TIME_HM },
	};
	bool first_pass;
	size_t i;

	memset(time, 0, sizeof(struct time));

	first_pass = true;

restart:
	for (i = 0; i < ARRAY_LEN(formats); i++) {
		struct tm tm;
		char *next;

		next = strptime(str, formats[i].fmt, &tm);
		if (next == NULL)
			continue;

		/* what is after the parsed chars? */
		switch (*next) {
			case ' ':
			case '\t':
			case '\0':
				/* end or whitespace - ok */
				next = find_nonws(next);
				break;
			default:
				/* garbage at the end - ignore format */
				continue;
		}

		/* parsed ok, but is it valid? */
		if ((formats[i].have & TIME_PART_DATE) &&
		    !valid_date(tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday))
			continue; /* impossible date */
		if ((formats[i].have & TIME_PART_TIME_HM) &&
		    ((tm.tm_hour > 23) || (tm.tm_min > 59)))
			continue; /* impossible hours/minutes */
		if ((formats[i].have & TIME_PART_TIME_S) && (tm.tm_sec > 60))
			continue; /* impossible seconds */

		/* did we get two times or two dates? */
		if (formats[i].have & time->have)
			return -EINVAL;

		if (formats[i].have & TIME_PART_DATE) {
			time->year = tm.tm_year + 1900;
			time->month = tm.tm_mon + 1;
			time->day = tm.tm_mday;
			time->have |= TIME_PART_DATE;
		}

		if (formats[i].have & TIME_PART_TIME_HM) {
			time->hour = tm.tm_hour;
			time->min = tm.tm_min;
			time->have &= ~TIME_PART_TIME_S; /* clear the seconds */
			time->have |= TIME_PART_TIME_HM;
		}

		if (formats[i].have & TIME_PART_TIME_S) {
			time->sec = tm.tm_sec;
			time->have |= TIME_PART_TIME_S;
		}

		/* if there is more to parse, do a second pass */
		if (first_pass && *next) {
			str = next;
			first_pass = false;
			goto restart;
		}

		return 0;
	}

	return -EINVAL;
}

int qso_set_datetime(struct qso *qso, bool start, const char *str)
{
	struct time *ts = start ? &qso->start : &qso->end;
	struct time tmp;
	char *raw;
	int ret;

	if (!str) {
		time_set_now(ts); /* set it to the current date & time */

		return 0;
	}

	raw = alloca(strlen(str) + 1);
	strcpy(raw, str);

	ret = __qso_set_datetime(&tmp, raw);
	if (ret)
		return ret;

	/* fill in qso time */

	if (tmp.have & TIME_PART_DATE) {
		ts->year = tmp.year;
		ts->month = tmp.month;
		ts->day = tmp.day;
		ts->have |= TIME_PART_DATE;
	}

	if (tmp.have & TIME_PART_TIME_HM) {
		ts->hour = tmp.hour;
		ts->min = tmp.min;
		ts->have &= ~TIME_PART_TIME_S; /* clear the seconds */
		ts->have |= TIME_PART_TIME_HM;
	}

	if (tmp.have & TIME_PART_TIME_S) {
		ts->sec = tmp.sec;
		ts->have |= TIME_PART_TIME_S;
	}

	return 0;
}

int qso_set_datetime_val(struct qso *qso, bool start, struct val *val)
{
	int ret;

	if (!val)
		return qso_set_datetime(qso, start, NULL);

	if (val->type != VT_STR) {
		val_putref(val);
		return -EINVAL;
	}

	ret = qso_set_datetime(qso, start, val_cstr(val));
	val_putref(val);
	return ret;
}

void qso_clear_datetime(struct qso *qso, bool start)
{
	struct time *ts = start ? &qso->start : &qso->end;

	ts->have = 0;
}

/* comment */
int qso_set_comment(struct qso *qso, const char *comment)
{
	struct str *str;

	str = str_dup(comment);
	if (IS_ERR(str))
		return PTR_ERR(str);

	return qso_set_comment_str(qso, str);
}

int qso_set_comment_str(struct qso *qso, struct str *comment)
{
	val_putref(qso->comment);
	qso->comment = str_cast_to_val(comment);

	return 0;
}

int qso_set_comment_val(struct qso *qso, struct val *comment)
{
	if (comment && (comment->type != VT_STR)) {
		val_putref(comment);
		return -EINVAL;
	} else {
		return qso_set_comment_str(qso, val_cast_to_str(comment));
	}
}

void qso_clear_comment(struct qso *qso)
{
	val_putref(qso->comment);
	qso->comment = NULL;
}

struct str *qso_get_comment(const struct qso *qso)
{
	return val_getref_str(qso->comment);
}

struct val *qso_get_comment_val(const struct qso *qso)
{
	return val_getref(qso->comment);
}

/*
 * QSO side functions
 */

/*
 * We only care to generate the _val get functions for when the short
 * function name returns a struct str *, but that's too annoying to
 * implement.  So, we *always* generate the _val get functions but for the
 * non-str case alias it to the short name.  Thereby, we accomplish the same
 * thing with not much impact.
 *
 * There are no prototypes for the _val get functions that are aliased,
 * because they don't add anything to the API.
 */

/* Generate the get val, set val, and clear functions. */
#define GEN_GSC(fxn, member, vtype)					\
struct val *qso_get_##fxn##_val(const struct qso_side *side)		\
{									\
	return val_getref(side->member);				\
}									\
int qso_set_##fxn##_val(struct qso_side *side, struct val *val)		\
{									\
	/* convert VT_NULL to actuall NULL pointer */			\
	if (val && (val->type == VT_NULL)) {				\
		val_putref(val);					\
		val = NULL;						\
	}								\
									\
	if (val && (val->type != vtype)) {				\
		val_putref(val);					\
		return -EINVAL;						\
	} else {							\
		val_putref(side->member);				\
		side->member = val;					\
		return 0;						\
	}								\
}									\
void qso_clear_##fxn(struct qso_side *side)				\
{									\
	val_putref(side->member);					\
	side->member = NULL;						\
}

/*
 * The generated qso_get_foo returns struct str *.
 */
#define GET_STR(fxn, member)						\
struct str *qso_get_##fxn(const struct qso_side *side)			\
{									\
	return val_getref_str(side->member);				\
}

/*
 * Generate qso_set_foo{,_str}.  The generated qso_set_foo takes a const
 * char *, while the generated qso_set_foo_str takes a struct str *.
 *
 * The former str_dups the argument and then passes it to qso_set_foo_val.
 * The latter casts the argument and passes it to qso_set_foo_val.
 */
#define SET_STR(fxn, member)						\
int qso_set_##fxn(struct qso_side *side, const char *cstr)		\
{									\
	struct str *str;						\
									\
	str = str_dup(cstr);						\
	if (IS_ERR(str))						\
		return PTR_ERR(str);					\
									\
	return qso_set_##fxn##_val(side, str_cast_to_val(str));		\
}									\
int qso_set_##fxn##_str(struct qso_side *side, struct str *str)		\
{									\
	return qso_set_##fxn##_val(side, str_cast_to_val(str));		\
}

/*
 * Generate an alias.  We use weak symbols if we can and fall back to plain
 * ol' wrapper functions.
 */
#ifdef HAVE_PRAGMA_WEAK
#define ALIAS_STRINGIFY(x) #x
#define GEN_ALIAS(name)							\
       _Pragma(ALIAS_STRINGIFY(weak qso_get_##name = qso_get_##name##_val))
#else
#define GEN_ALIAS(name)							\
struct val *qso_get_##name(const struct qso_side *side)			\
{									\
       return qso_get_##name##_val(side);				\
}
#endif

/* station call */
int qso_set_station_call(struct qso_side *side, const char *call)
{
	char ucall[strlen(call) + 1];

	uppercase_copy(ucall, call);

	return qso_set_station_call_val(side, VAL_DUP_STR(ucall));
}

GEN_GSC(station_call, station_call, VT_STR);
GET_STR(station_call, station_call);

/* operator call */
int qso_set_operator_call(struct qso_side *side, const char *call)
{
	char ucall[strlen(call) + 1];

	uppercase_copy(ucall, call);

	return qso_set_operator_call_val(side, VAL_DUP_STR(ucall));
}

GEN_GSC(operator_call, operator_call, VT_STR);
GET_STR(operator_call, operator_call);

/* freq */
int qso_set_freq(struct qso_side *side, const char *str)
{
	int64_t freq;
	int ret;

	ret = extract_float_as_int64(str, 1000000, &freq);
	if (ret)
		return ret;

	val_putref(side->freq);

	if (!freq)
		side->freq = NULL;
	else
		side->freq = VAL_ALLOC_INT(freq);

	return 0;
}

GEN_GSC(freq, freq, VT_INT);
GEN_ALIAS(freq)

/* grid */
int qso_set_grid(struct qso_side *side, const char *grid)
{
	char ngrid[strlen(grid) + 1];

	/* 24th IARU Region 1 General Conference said to use all uppercase */
	uppercase_copy(ngrid, grid);

	return qso_set_grid_val(side, VAL_DUP_STR(ngrid));
}

GEN_GSC(grid, grid, VT_STR);
GET_STR(grid, grid);

/* power */
int qso_set_power(struct qso_side *side, uint64_t milliwatts)
{
	val_putref(side->power);

	if (!milliwatts)
		side->power = NULL;
	else
		side->power = VAL_ALLOC_INT(milliwatts);

	return 0;
}

int qso_set_power_user(struct qso_side *side, const char *power)
{
	int64_t milliwatts;
	int ret;

	ret = extract_float_as_int64(power, 1000, &milliwatts);
	if (ret)
		return ret;

	return qso_set_power(side, milliwatts);
}

GEN_GSC(power, power, VT_INT);
GEN_ALIAS(power)

/* mode */
int qso_set_mode(struct qso_side *side, const char *mode)
{
	static struct str common_modes[] = {
		STR_STATIC_INITIALIZER("CW"),
		STR_STATIC_INITIALIZER("AM"),
		STR_STATIC_INITIALIZER("FM"),
		STR_STATIC_INITIALIZER("SSB"),
		STR_STATIC_INITIALIZER("USB"),
		STR_STATIC_INITIALIZER("LSB"),
		STR_STATIC_INITIALIZER("RTTY"),
		STR_STATIC_INITIALIZER("FT4"),
		STR_STATIC_INITIALIZER("FT8"),
		STR_STATIC_INITIALIZER("PSK31"),
		STR_STATIC_INITIALIZER("QPSK31"),
		STR_STATIC_INITIALIZER("OLIVIA"),
		STR_STATIC_INITIALIZER("DOMEX8"),
		STR_STATIC_INITIALIZER("THOR4"),
	};
	size_t i;

	if (side->mode &&
	    !strcmp(val_cstr(side->mode), mode))
		return 0; /* no change */

	for (i = 0; i < ARRAY_LEN(common_modes); i++) {
		if (strcasecmp(str_cstr(&common_modes[i]), mode))
			continue;

		return qso_set_mode_val(side, str_cast_to_val(&common_modes[i]));
	}

	/* non-common mode, use it as-is */
	return qso_set_mode_val(side, VAL_DUP_STR(mode));
}

GEN_GSC(mode, mode, VT_STR);
GET_STR(mode, mode);

/* rst */
int qso_set_rst(struct qso_side *side, const char *rst)
{
	return qso_set_rst_val(side, VAL_DUP_STR(rst));
}

GEN_GSC(rst, rst, VT_STR);
GET_STR(rst, rst);

/* name */
GEN_GSC(name, name, VT_STR);
GET_STR(name, name);
SET_STR(name, name);

/* qth */
GEN_GSC(qth, qth, VT_STR);
GET_STR(qth, qth);
SET_STR(qth, qth);

/* country */
GEN_GSC(country, country, VT_STR);
GET_STR(country, country);
SET_STR(country, country);

/* state */
GEN_GSC(state, state, VT_STR);
GET_STR(state, state);
SET_STR(state, state);

/* county */
GEN_GSC(county, county, VT_STR);
GET_STR(county, county);
SET_STR(county, county);

/* continent */
int qso_set_continent(struct qso_side *side, const char *zone)
{
	static struct str continents[] = {
		STR_STATIC_INITIALIZER("AF"),
		STR_STATIC_INITIALIZER("AN"),
		STR_STATIC_INITIALIZER("AS"),
		STR_STATIC_INITIALIZER("EU"),
		STR_STATIC_INITIALIZER("NA"),
		STR_STATIC_INITIALIZER("OC"),
		STR_STATIC_INITIALIZER("SA"),
	};
	size_t i;

	if (!zone)
		goto out;

	for (i = 0; i < ARRAY_LEN(continents); i++) {
		if (strcasecmp(zone, str_cstr(&continents[i])))
			continue;

		/* a real continent */
		return qso_set_continent_val(side,
					     str_cast_to_val(&continents[i]));
	}

out:
	/* a non-existant continent - clear it */
	return qso_set_continent_val(side, NULL);
}

GEN_GSC(continent, continent, VT_STR);
GET_STR(continent, continent);

/* dxcc */
int qso_set_zone_dxcc(struct qso_side *side, uint64_t zone)
{
	/*
	 * Note: in ADIF, code==0 indicates that station is known not to be
	 * within a DXCC entitity
	 */
	return qso_set_zone_dxcc_val(side, VAL_ALLOC_INT(zone));
}

int qso_set_zone_dxcc_user(struct qso_side *side, const char *zone)
{
	uint64_t tmp;
	int ret;

	ret = str2u64(zone, &tmp);
	if (!ret)
		ret = qso_set_zone_dxcc(side, tmp);

	return ret;
}

GEN_GSC(zone_dxcc, dxcc, VT_INT);
GEN_ALIAS(zone_dxcc)

/* itu zone */
int qso_set_zone_itu(struct qso_side *side, uint64_t zone)
{
	/* valid zones are [1,90], 0 clears it */
	if (!zone) {
		qso_clear_zone_itu(side);
		return 0;
	} else if ((zone >= 1) && (zone <= 90)) {
		return qso_set_zone_itu_val(side, VAL_ALLOC_INT(zone));
	} else {
		return -ERANGE;
	}
}

int qso_set_zone_itu_user(struct qso_side *side, const char *zone)
{
	uint64_t tmp;
	int ret;

	ret = str2u64(zone, &tmp);
	if (!ret)
		ret = qso_set_zone_itu(side, tmp);

	return ret;
}

GEN_GSC(zone_itu, itu, VT_INT);
GEN_ALIAS(zone_itu)

/* cq zone */
int qso_set_zone_cqz(struct qso_side *side, uint64_t zone)
{
	/* valid zones are [1,40], 0 clears it */
	if (!zone) {
		qso_clear_zone_cqz(side);
		return 0;
	} else if ((zone >= 1) && (zone <= 40)) {
		return qso_set_zone_cqz_val(side, VAL_ALLOC_INT(zone));
	} else {
		return -ERANGE;
	}
}

int qso_set_zone_cqz_user(struct qso_side *side, const char *zone)
{
	uint64_t tmp;
	int ret;

	ret = str2u64(zone, &tmp);
	if (!ret)
		ret = qso_set_zone_cqz(side, tmp);

	return ret;
}

GEN_GSC(zone_cqz, cqz, VT_INT);
GEN_ALIAS(zone_cqz)

/* seq */
int qso_set_seq(struct qso_side *side, uint64_t seq)
{
	/* Note: 0 is a valid sequence number */
	return qso_set_seq_val(side, VAL_ALLOC_INT(seq));
}

int qso_set_seq_user(struct qso_side *side, const char *seq)
{
	uint64_t tmp;
	int ret;

	ret = str2u64(seq, &tmp);
	if (!ret)
		ret = qso_set_seq(side, tmp);

	return ret;
}

GEN_GSC(seq, seq, VT_INT);
GEN_ALIAS(seq)

/* band - derived from freq, no set/clear */
struct val *qso_get_band(const struct qso_side *side)
{
	if (!side->freq)
		return NULL;

	return bands_lookup_lambda_val_by_freq(side->freq->i);
}
