/*
 * Copyright (c) 2020-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/cstr.h>
#include <jeffpc/time.h>

#include <hlog/time.h>

#define TIME_CMP_HELPER(a, b, member) \
	do { \
		if ((a)->member < (b)->member) \
			return -1; \
		if ((a)->member > (b)->member) \
			return +1; \
	} while (0)

int time_cmp(const struct time *a, const struct time *b)
{
	if ((a->have & TIME_PART_DATE) &&
	    (b->have & TIME_PART_DATE)) {
		TIME_CMP_HELPER(a, b, year);
		TIME_CMP_HELPER(a, b, month);
		TIME_CMP_HELPER(a, b, day);
	} else if (a->have & TIME_PART_DATE) {
		return +1;
	} else if (b->have & TIME_PART_DATE) {
		return -1;
	} else {
		/* both mising date - equal so far */
	}

	if ((a->have & TIME_PART_TIME_HM) &&
	    (b->have & TIME_PART_TIME_HM)) {
		TIME_CMP_HELPER(a, b, hour);
		TIME_CMP_HELPER(a, b, min);
	} else if (a->have & TIME_PART_TIME_HM) {
		return +1;
	} else if (b->have & TIME_PART_TIME_HM) {
		return -1;
	} else {
		/* both mising time - equal so far */
	}

	if ((a->have & TIME_PART_TIME_S) &&
	    (b->have & TIME_PART_TIME_S)) {
		TIME_CMP_HELPER(a, b, sec);
	} else if (a->have & TIME_PART_TIME_S) {
		return +1;
	} else if (b->have & TIME_PART_TIME_S) {
		return -1;
	}

	return 0; /* both missing time - actually equal */
}

void time_set_now(struct time *ts)
{
	struct tm tm;
	time_t now;

	now = time(NULL);
	gmtime_r(&now, &tm);

	ts->year = tm.tm_year + 1900;
	ts->month = tm.tm_mon + 1;
	ts->day = tm.tm_mday;
	ts->hour = tm.tm_hour;
	ts->min = tm.tm_min;
	ts->sec = tm.tm_sec;

	ts->have = TIME_PART_DATE | TIME_PART_TIME_HMS;
}

int time_parse(struct time *time, const char *str, bool date)
{
	uint64_t tmp;
	int ret;

	memset(time, 0, sizeof(struct time));

	ret = str2u64(str, &tmp);
	if (ret)
		return ret;

	switch (strlen(str)) {
		case 4:
			if (date)
				return -ERANGE;

			time->hour = tmp / 100;
			time->min = tmp % 100;
			time->have = TIME_PART_TIME_HM;

			if (!valid_time(time->hour, time->min, 0))
				return -ERANGE;
			break;
		case 6:
			if (date)
				return -ERANGE;

			time->hour = tmp / 10000;
			time->min = (tmp / 100) % 100;
			time->sec = tmp % 100;
			time->have = TIME_PART_TIME_HMS;

			if (!valid_time(time->hour, time->min, time->sec))
				return -ERANGE;
			break;
		case 8:
			if (!date)
				return -ERANGE;

			time->year = tmp / 10000;
			time->month = (tmp / 100) % 100;
			time->day = tmp % 100;
			time->have = TIME_PART_DATE;

			if (!valid_date(time->year, time->month, time->day))
				return -ERANGE;
			break;
		default:
			return -ERANGE;
	}

	return 0;
}

bool time_snprint_date(char *buf, size_t bufsize, const struct time *t)
{
	if (!t || !(t->have & TIME_PART_DATE)) {
		strcpy_safe(buf, "????""-??""-??", bufsize);

		return false;
	} else {
		snprintf(buf, bufsize, "%04u-%02u-%02u", t->year, t->month,
			 t->day);

		return true;
	}
}

bool time_snprint_time(char *buf, size_t bufsize, const struct time *t)
{
	if (!t || !(t->have & TIME_PART_TIME_HM)) {
		strcpy_safe(buf, "??:??:??", bufsize);

		return false;
	} else {
		if (t->have & TIME_PART_TIME_S)
			snprintf(buf, bufsize, "%02u:%02u:%02u",
				 t->hour, t->min, t->sec);
		else
			snprintf(buf, bufsize, "%02u:%02u:__",
				 t->hour, t->min);

		return true;
	}
}

/* write out a ISO 8601-esque timestamp */
bool time_snprint_iso8601(char *buf, size_t bufsize, const struct time *t)
{
	if ((t->have & TIME_PART_DATE) && (t->have & TIME_PART_TIME_HM) &&
	    (t->have & TIME_PART_TIME_S))
		snprintf(buf, bufsize, "%04u-%02u-%02uT%02u:%02u:%02uZ",
			 t->year, t->month, t->day, t->hour, t->min, t->sec);
	else if ((t->have & TIME_PART_DATE) && (t->have & TIME_PART_TIME_HM))
		snprintf(buf, bufsize, "%04u-%02u-%02uT%02u:%02uZ",
			 t->year, t->month, t->day, t->hour, t->min);
	else if (t->have & TIME_PART_DATE)
		snprintf(buf, bufsize, "%04u-%02u-%02u",
			 t->year, t->month, t->day);
	else if ((t->have & TIME_PART_TIME_HM) && (t->have & TIME_PART_TIME_S))
		snprintf(buf, bufsize, "%02u:%02u:%02uZ",
			 t->hour, t->min, t->sec);
	else if (t->have & TIME_PART_TIME_HM)
		snprintf(buf, bufsize, "%02u:%02uZ",
			 t->hour, t->min);
	else
		snprintf(buf, bufsize, ""); /* nothing or seconds only */

	return (t->have && (t->have != TIME_PART_TIME_S));
}

bool valid_date(uint32_t y, uint32_t m, uint32_t d)
{
	static const uint8_t month_days[12] = {
		31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31,
	};

	if ((y < 1800) || !m || !d)
		return false;

	if (d > month_days[m - 1])
		return false;

	/* leap year - all good */
	if (((y % 4) == 0) &&
	    (((y % 400) == 0) || ((y % 100) != 0)))
		return true;

	/* non-leap year - Feb 29 is bad */
	return (m != 2) || (d < 29);
}

bool valid_time(uint32_t h, uint32_t m, uint32_t s)
{
	if (h >= 24)
		return false;
	if (m >= 60)
		return false;

	if ((h == 23) && (m == 59))
		return s <= 60; /* leap second */
	else
		return s < 60; /* non-leap second */
}
