/*
 * Copyright (c) 2021-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * Wrappers around hamlib for ease of use
 */

#include <hamlib/rig.h>

#include <jeffpc/types.h>
#include <jeffpc/error.h>
#include <jeffpc/mem.h>
#include <jeffpc/synch.h>

#include <hlog/config.h>

#include "hamlib.h"

static LOCK_CLASS(lc_rig);

struct hrig {
	struct lock lock;
	RIG *rig;

	bool open;

	struct {
		bool no_txpower;
	} quirks;

	/* cached values */
	vfo_t rx_vfo;
	vfo_t tx_vfo;
	bool split;
};

static int hamlib_err2errno(int e)
{
#define NOT_YET(e) \
	case e: \
		panic("Hamlib error code %s not yet converted by %s", \
		      #e, __func__)

	switch ((enum rig_errcode_e) -e) {
		case RIG_OK:
			return 0;
		case RIG_EINVAL:
			return -EINVAL;
		NOT_YET(RIG_ECONF);
		case RIG_ENOMEM:
			return -ENOMEM;
		case RIG_ENIMPL:
			return -ENOTSUP;
		case RIG_ETIMEOUT:
			return -ETIMEDOUT;
		case RIG_EIO:
			return -EIO;
		NOT_YET(RIG_EINTERNAL);
		case RIG_EPROTO:
			return -EPROTO;
		NOT_YET(RIG_ERJCTED);
		NOT_YET(RIG_ETRUNC);
		NOT_YET(RIG_ENAVAIL);
		NOT_YET(RIG_ENTARGET);
		NOT_YET(RIG_BUSERROR);
		NOT_YET(RIG_BUSBUSY);
		NOT_YET(RIG_EARG);
		NOT_YET(RIG_EVFO);
		case RIG_EDOM:
			return -EDOM;
#ifdef HLOG_HAVE_RIG_EDEPRECATED
		case RIG_EDEPRECATED:
			return -ENOTSUP;
#endif
#ifdef HLOG_HAVE_RIG_ESECURITY
		case RIG_ESECURITY:
			return -EPERM;
#endif
	}
#undef NOT_YET

	panic("Unknown hamlib error code: %u", e);
}

#define MAP()							\
	ENTRY(HRIG_MODE_AM,       RIG_MODE_AM)			\
	ENTRY(HRIG_MODE_C4FM,     RIG_MODE_C4FM)		\
	ENTRY(HRIG_MODE_FM,       RIG_MODE_FM)			\
	ENTRY(HRIG_MODE_FM_DATA,  RIG_MODE_PKTFM)		\
	ENTRY(HRIG_MODE_LSB,      RIG_MODE_LSB)			\
	ENTRY(HRIG_MODE_LSB_CW,   RIG_MODE_CWR)			\
	ENTRY(HRIG_MODE_LSB_DATA, RIG_MODE_PKTLSB)		\
	ENTRY(HRIG_MODE_LSB_RTTY, RIG_MODE_RTTYR)		\
	ENTRY(HRIG_MODE_USB,      RIG_MODE_USB)			\
	ENTRY(HRIG_MODE_USB_CW,   RIG_MODE_CW)			\
	ENTRY(HRIG_MODE_USB_DATA, RIG_MODE_PKTUSB)		\
	ENTRY(HRIG_MODE_USB_RTTY, RIG_MODE_RTTY)

static int hamlib_mode2hmode(rmode_t mode)
{
	switch (mode) {
#define ENTRY(hmode, rmode)	case rmode: return hmode;
		MAP()
#undef ENTRY
	}

	return HRIG_MODE_UNKNOWN;
}

static rmode_t hamlib_hmode2mode(enum hrig_mode mode)
{
	switch (mode) {
		case HRIG_MODE_UNKNOWN:
			break;

#define ENTRY(hmode, rmode)	case hmode: return rmode;
		MAP()
#undef ENTRY
	}

	return RIG_MODE_NONE;
}

const char *hamlib_mode_name(enum hrig_mode mode)
{
	switch (mode) {
		case HRIG_MODE_UNKNOWN: return "<unknown>";
		case HRIG_MODE_AM: return "AM";
		case HRIG_MODE_C4FM: return "C4FM";
		case HRIG_MODE_FM: return "FM";
		case HRIG_MODE_FM_DATA: return "DATA-FM";
		case HRIG_MODE_LSB: return "LSB";
		case HRIG_MODE_LSB_CW: return "CW-L";
		case HRIG_MODE_LSB_DATA: return "DATA-L";
		case HRIG_MODE_LSB_RTTY: return "RTTY-L";
		case HRIG_MODE_USB: return "USB";
		case HRIG_MODE_USB_CW: return "CW-U";
		case HRIG_MODE_USB_DATA: return "DATA-U";
		case HRIG_MODE_USB_RTTY: return "RTTY-U";
	}

	panic("Invalid hrig_mode value: %u", mode);
}

struct hrig *hamlib_alloc(uint32_t model, const char *dev)
{
	struct hrig *rig;
	int ret;

	/* avoid spamming the console with pointless messages */
	rig_set_debug_level(RIG_DEBUG_NONE);

	rig = malloc(sizeof(struct hrig));
	if (!rig)
		return ERR_PTR(-ENOMEM);

	MXINIT(&rig->lock, &lc_rig);
	rig->open = false;

	rig->rig = rig_init(model);
	if (!rig->rig) {
		cmn_err(CE_ERROR, "Failed to initialize rig");
		ret = -ENOMEM;
		goto err_rig;
	}

	/* TODO: is there a proper way to set the path? */
	strncpy(rig->rig->state.rigport.pathname, dev,
		sizeof(rig->rig->state.rigport.pathname) - 1);

	if (model == RIG_MODEL_NETRIGCTL) {
		rig->quirks.no_txpower = true;
		cmn_err(CE_WARN, "Hamlib 4.3.1 and older has buggy netrigctl");
		cmn_err(CE_WARN, "Disabling power announcements");
	}

	return rig;

err_rig:
	free(rig);

	return ERR_PTR(ret);
}

void hamlib_free(struct hrig *rig)
{
	rig_close(rig->rig);

	MXDESTROY(&rig->lock);

	free(rig);
}

int hamlib_open(struct hrig *rig)
{
	int ret;

	MXLOCK(&rig->lock);

	if (rig->open) {
		ret = -EBUSY;
		goto err;
	}

	ret = hamlib_err2errno(rig_open(rig->rig));

	rig->open = (ret == 0);

err:
	MXUNLOCK(&rig->lock);

	return ret;
}

int hamlib_close(struct hrig *rig)
{
	int ret;

	MXLOCK(&rig->lock);

	if (!rig->open) {
		ret = -ENOTCONN;
		goto err;
	}

	ret = hamlib_err2errno(rig_close(rig->rig));

	rig->open = (rig == 0);

err:
	MXUNLOCK(&rig->lock);

	return ret;
}

static int hamlib_refresh_cache(struct hrig *rig)
{
	split_t split;
	int ret;

	if (!rig->open)
		return -ENOTCONN;

	ret = rig_get_vfo(rig->rig, &rig->rx_vfo);
	if (ret)
		goto err;

	ret = rig_get_split_vfo(rig->rig, rig->rx_vfo, &split, &rig->tx_vfo);
	if (ret)
		goto err;

	rig->split = split == RIG_SPLIT_ON;

	return 0;

err:
	return hamlib_err2errno(ret);
}

static int hamlib_get_state_unlocked(struct hrig *rig, struct hrig_state *state)
{
	rmode_t tx_mode, rx_mode;
	freq_t tx_freq, rx_freq;
	unsigned int power;
	pbwidth_t width;
	ptt_t ptt;
	int ret;

	/*
	 *  1. get tx freq & mode
	 *     split=0: \get_freq \get_mode
	 *     split=1: \get_freq \get_mode
	 *  2. get rx freq & mode
	 *     split=0: \get_freq \get_mode
	 *     split=1: \get_split_freq \get_split_mode
	 *  3. get tx power
	 *     \get_level RFPOWER
	 *     \power2mW <0.25> <freq> <mode>
	 */

	ret = hamlib_refresh_cache(rig);
	if (ret)
		return ret;

	/* get rx freq and mode */
	ret = rig_get_freq(rig->rig, rig->rx_vfo, &rx_freq);
	if (ret)
		goto err;

	ret = rig_get_mode(rig->rig, rig->rx_vfo, &rx_mode, &width);
	if (ret)
		goto err;

	/* get tx freq and mode */
	if (rig->split) {
		ret = rig_get_split_freq_mode(rig->rig, rig->tx_vfo,
					      &tx_freq, &tx_mode,
					      &width);
		if (ret)
			goto err;
	} else {
		tx_freq = rx_freq;
		tx_mode = rx_mode;
	}

	/*
	 * get tx power
	 *
	 * NOTE: Older versions of hamlib have a buggy netrigctl that may
	 * segfault due to a buffer overflow when calling rig_power2mW, so
	 * we don't risk it and for now send a zero which means "no power
	 * information available".
	 */
	if (rig->quirks.no_txpower) {
		power = 0;
	} else {
		value_t val;

		ret = rig_get_level(rig->rig, RIG_VFO_CURR, RIG_LEVEL_RFPOWER,
				    &val);
		if (ret)
			goto err;

		ret = rig_power2mW(rig->rig, &power, val.f, tx_freq, tx_mode);
		if (ret)
			goto err;
	}

	/* get ptt */
	ret = rig_get_ptt(rig->rig, rig->rx_vfo, &ptt);
	if (ret)
		goto err;

	state->tx_freq = tx_freq;
	state->tx_mode = hamlib_mode2hmode(tx_mode);
	state->rx_freq = rx_freq;
	state->rx_mode = hamlib_mode2hmode(rx_mode);
	state->power = power;
	state->split = rig->split;
	state->ptt = ptt != RIG_PTT_OFF;

	return 0;

err:
	return hamlib_err2errno(ret);
}

int hamlib_get_state(struct hrig *rig, struct hrig_state *state)
{
	int ret;

	MXLOCK(&rig->lock);
	ret = hamlib_get_state_unlocked(rig, state);
	MXUNLOCK(&rig->lock);

	return ret;
}

int hamlib_get_ptt(struct hrig *rig, bool *on)
{
	return -ENOTSUP;
}

static int hamlib_set_ptt_unlocked(struct hrig *rig, bool on)
{
	int ret;

	ret = hamlib_refresh_cache(rig);
	if (ret)
		return ret;

	ret = rig_set_ptt(rig->rig, rig->tx_vfo,
			  on ? RIG_PTT_ON : RIG_PTT_OFF);

	if (ret != RIG_OK)
		cmn_err(CE_WARN, "hamlib: set ptt to %s: %s",
			on ? "tx" : "rx", rigerror(ret));

	return hamlib_err2errno(ret);
}

int hamlib_set_ptt(struct hrig *rig, bool on)
{
	int ret;

	MXLOCK(&rig->lock);
	ret = hamlib_set_ptt_unlocked(rig, on);
	MXUNLOCK(&rig->lock);

	return ret;
}

int hamlib_get_mode(struct hrig *rig, enum hrig_mode *mode)
{
	return -ENOTSUP;
}

static int hamlib_set_mode_unlocked(struct hrig *rig, enum hrig_mode mode,
				    bool rx_if_split)
{
	const pbwidth_t width = 0;
	rmode_t rmode;
	int ret;

	rmode = hamlib_hmode2mode(mode);
	if (rmode == RIG_MODE_NONE)
		return -EINVAL;

	ret = hamlib_refresh_cache(rig);
	if (ret)
		return ret;

	if (!rig->split || rx_if_split)
		ret = rig_set_mode(rig->rig, rig->rx_vfo, rmode, width);
	else
		ret = rig_set_split_mode(rig->rig, rig->tx_vfo, rmode, width);

	if (ret != RIG_OK)
		cmn_err(CE_WARN, "hamlib: set %s mode to %s: %s",
			!rig->split ? "tx+rx" : (rx_if_split ? "rx" : "tx"),
			hamlib_mode_name(mode), rigerror(ret));

	return hamlib_err2errno(ret);
}

int hamlib_set_mode(struct hrig *rig, enum hrig_mode mode, bool rx_if_split)
{
	int ret;

	MXLOCK(&rig->lock);
	ret = hamlib_set_mode_unlocked(rig, mode, rx_if_split);
	MXUNLOCK(&rig->lock);

	return ret;
}

int hamlib_get_freq(struct hrig *rig, uint64_t *freq)
{
	return -ENOTSUP;
}

static int hamlib_set_freq_unlocked(struct hrig *rig, uint64_t freq,
				    bool rx_if_split)
{
	int ret;

	ret = hamlib_refresh_cache(rig);
	if (ret)
		return ret;

	if (!rig->split || rx_if_split)
		ret = rig_set_freq(rig->rig, rig->rx_vfo, freq);
	else
		ret = rig_set_split_freq(rig->rig, rig->tx_vfo, freq);

	if (ret != RIG_OK)
		cmn_err(CE_WARN, "hamlib: set %s freq to %"PRIu64": %s",
			!rig->split ? "tx+rx" : (rx_if_split ? "rx" : "tx"),
			freq, rigerror(ret));

	return hamlib_err2errno(ret);
}

int hamlib_set_freq(struct hrig *rig, uint64_t freq, bool rx_if_split)
{
	int ret;

	MXLOCK(&rig->lock);
	ret = hamlib_set_freq_unlocked(rig, freq, rx_if_split);
	MXUNLOCK(&rig->lock);

	return ret;
}
