/*
 * Copyright (c) 2020-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/version.h>
#include <jeffpc/error.h>
#include <jeffpc/time.h>
#include <jeffpc/val.h>
#include <jeffpc/io.h>
#include <jeffpc/unicode.h>
#include <jeffpc/synch.h>
#include <jeffpc/cstr.h>
#include <jeffpc/mem.h>

#include <ctype.h>
#include <unistd.h>
#include <locale.h>
#include <math.h>
#include <spawn.h>

#include <hlog/version.h>
#include <hlog/qso.h>
#include <hlog/qso_line.h>
#include <hlog/adif.h>
#include <hlog/bands.h>
#include <hlog/index.h>
#include <hlog/country.h>
#include <hlog/maidenhead.h>
#include <hlog/util.h>
#include <hlog-lua/xlua.h>
#include <hlog-rpc/announce.h>
#include <hlog-rpc/announce-sub.h>

#include "common/config.h"
#include "common/databases.h"
#include "common/recent.h"
#include "common/status.h"
#include "common/xcurses.h"
#include "common/xform.h"

#include "contest/contest.h"
#include "contest/script.h"

#define LEFT_PANEL_WIDTH	100

#define HEADER_WIN_HEIGHT	1
#define HEADER_WIN_WIDTH	LEFT_PANEL_WIDTH
#define HEADER_WIN_ROW		0
#define HEADER_WIN_COL		0
#define CALLOUT_WIN_HEIGHT	4
#define CALLOUT_WIN_WIDTH	35
#define CALLOUT_WIN_ROW		2
#define CALLOUT_WIN_COL		(LEFT_PANEL_WIDTH - CALLOUT_WIN_WIDTH)
#define SUPERCHECK_WIN_HEIGHT	7
#define SUPERCHECK_WIN_WIDTH	(CALLOUT_WIN_COL - 1)
#define SUPERCHECK_WIN_ROW	2
#define SUPERCHECK_WIN_COL	0
#define WORK_WIN_HEIGHT		2
#define WORK_WIN_WIDTH		SUPERCHECK_WIN_WIDTH
#define WORK_WIN_ROW		(SUPERCHECK_WIN_ROW + SUPERCHECK_WIN_HEIGHT)
#define WORK_WIN_COL		0
#define CURRENT_WIN_HEIGHT	7
#define CURRENT_WIN_WIDTH	LEFT_PANEL_WIDTH
#define CURRENT_WIN_ROW		(LINES - CURRENT_WIN_HEIGHT)
#define CURRENT_WIN_COL		0
#define STATUS_WIN_HEIGHT	1
#define STATUS_WIN_WIDTH	CURRENT_WIN_WIDTH
#define STATUS_WIN_ROW		(WORK_WIN_ROW + WORK_WIN_HEIGHT)
#define STATUS_WIN_COL		0
#define RECENT_WORK_WIN_WIDTH	CURRENT_WIN_WIDTH
#define RECENT_WORK_WIN_ROW	(STATUS_WIN_ROW + STATUS_WIN_HEIGHT)
#define RECENT_WORK_WIN_COL	0
#define RECENT_WORK_WIN_HEIGHT	5
#define RECENT_WIN_WIDTH	CURRENT_WIN_WIDTH
#define RECENT_WIN_ROW		(RECENT_WORK_WIN_ROW + RECENT_WORK_WIN_HEIGHT + 1)
#define RECENT_WIN_COL		0
#define RECENT_WIN_HEIGHT	(CURRENT_WIN_ROW - RECENT_WIN_ROW)

#define COLOR_HINT_BLACK	1
#define COLOR_HINT_RED		2
#define COLOR_HINT_GREEN	3
#define COLOR_HINT_YELLOW	4
#define COLOR_HINT_BLUE		5
#define COLOR_HINT_MAGENTA	6
#define COLOR_HINT_CYAN		7
#define COLOR_HINT_WHITE	8

#define ERROR_COLOR		COLOR_PAIR(COLOR_HINT_RED)
#define LABEL_XCOLOR		XFORM_COLOR_YELLOW

#define UNKNOWN_LOCATION_STR	"<unknown>"

static struct contest_params contest_params;

static const char *data_dir;
static struct qso *template;
static struct qso *contact;

static char location[MAIDENHEAD_MAX_STRING_LENGTH];
static struct lock location_lock;
static LOCK_CLASS(location_lc);

static struct rig_info {
	uint64_t tx_freq;
	uint64_t rx_freq;
	uint64_t power;
} rig_info;
static struct lock rig_lock;
static LOCK_CLASS(rig_lc);

static WINDOW *header_win; /* location & time */
static WINDOW *supercheck_win;
struct recent recent_work_list; /* current contact history */
struct recent recent_list; /* all history */

/* the currently active form */
static struct xform *active_form = &work;

static struct str label_de = STR_STATIC_INITIALIZER("de");
static struct str label_dx = STR_STATIC_INITIALIZER("dx");

struct field_info {
	struct val *(*get)(const struct qso_side *);
	enum {
		FI_SIDE_TX,
		FI_SIDE_RX,
	} what;
};

#define FIELD_INFO(lbl, name, fxn)					\
	static struct str label_##name = STR_STATIC_INITIALIZER(lbl);	\
	static struct field_info field_info_##name##_tx = {	\
		(fxn), FI_SIDE_TX,					\
	};								\
	static struct field_info field_info_##name##_rx = {	\
		(fxn), FI_SIDE_RX,					\
	}

FIELD_INFO("St. Call", station_call, qso_get_station_call_val);
FIELD_INFO("Op. Call", operator_call, qso_get_operator_call_val);
FIELD_INFO("Freq", freq, qso_get_freq);
FIELD_INFO("Mode", mode, qso_get_mode_val);
FIELD_INFO("RST", rst, qso_get_rst_val);
FIELD_INFO("Name", name, qso_get_name_val);
FIELD_INFO("Grid", grid, qso_get_grid_val);
FIELD_INFO("DXCC", zone_dxcc, qso_get_zone_dxcc);
FIELD_INFO("ITU", zone_itu, qso_get_zone_itu);
FIELD_INFO("CQZ", zone_cqz, qso_get_zone_cqz);
FIELD_INFO("Seq", seq, qso_get_seq);
FIELD_INFO("QTH", qth, qso_get_qth_val);
FIELD_INFO("St", state, qso_get_state_val);
FIELD_INFO("County", county, qso_get_county_val);
FIELD_INFO("Country", country, qso_get_country_val);
FIELD_INFO("Cont", continent, qso_get_continent_val);

/* attributes for labels */
#define LATTRS \
	{ \
		LABEL_XCOLOR, \
		XFORM_UNDERLINE_AUTO, \
	}

/* attributes for fields */
#define FATTRS \
	{ \
		XFORM_COLOR_NONE, \
		XFORM_UNDERLINE_AUTO, \
	}

/* all fields - for current form */
static const struct xfield_def cfields[] = {
	/* row 0 */
	{ XFIELD_TYPE_LABEL, 0,  3, 10, &label_station_call, LATTRS },
	{ XFIELD_TYPE_LABEL, 0, 14, 10, &label_operator_call, LATTRS },
	{ XFIELD_TYPE_LABEL, 0, 25, 10, &label_freq, LATTRS },
	/* TODO: power */
	{ XFIELD_TYPE_LABEL, 0, 36,  6, &label_mode, LATTRS },
	{ XFIELD_TYPE_LABEL, 0, 43,  3, &label_rst, LATTRS },
	{ XFIELD_TYPE_LABEL, 0, 47, 16, &label_name, LATTRS },
	{ XFIELD_TYPE_LABEL, 0, 64,  8, &label_grid, LATTRS },
	{ XFIELD_TYPE_LABEL, 0, 73,  3, &label_zone_dxcc, LATTRS },
	{ XFIELD_TYPE_LABEL, 0, 78,  2, &label_zone_itu, LATTRS },
	{ XFIELD_TYPE_LABEL, 0, 82,  2, &label_zone_cqz, LATTRS },
	{ XFIELD_TYPE_LABEL, 0, 86,  4, &label_seq, LATTRS },

	/* row 1 */
	{ XFIELD_TYPE_LABEL, 1,  0,  2, &label_de, LATTRS },
	{ XFIELD_TYPE_FIELD, 1,  3, 10, NULL, FATTRS, &field_info_station_call_tx },
	{ XFIELD_TYPE_FIELD, 1, 14, 10, NULL, FATTRS, &field_info_operator_call_tx },
	{ XFIELD_TYPE_FIELD, 1, 25, 10, NULL, FATTRS, &field_info_freq_tx },
	/* TODO: power */
	{ XFIELD_TYPE_FIELD, 1, 36,  6, NULL, FATTRS, &field_info_mode_tx },
	{ XFIELD_TYPE_FIELD, 1, 43,  3, NULL, FATTRS, &field_info_rst_tx },
	{ XFIELD_TYPE_FIELD, 1, 47, 16, NULL, FATTRS, &field_info_name_tx },
	{ XFIELD_TYPE_FIELD, 1, 64,  8, NULL, FATTRS, &field_info_grid_tx },
	{ XFIELD_TYPE_FIELD, 1, 73,  3, NULL, FATTRS, &field_info_zone_dxcc_tx },
	{ XFIELD_TYPE_FIELD, 1, 78,  2, NULL, FATTRS, &field_info_zone_itu_tx },
	{ XFIELD_TYPE_FIELD, 1, 82,  2, NULL, FATTRS, &field_info_zone_cqz_tx },
	{ XFIELD_TYPE_FIELD, 1, 86,  4, NULL, FATTRS, &field_info_seq_tx },

	/* row 2 */
	{ XFIELD_TYPE_LABEL, 2,  0,  2, &label_dx, LATTRS },
	{ XFIELD_TYPE_FIELD, 2,  3, 10, NULL, FATTRS, &field_info_station_call_rx },
	{ XFIELD_TYPE_FIELD, 2, 14, 10, NULL, FATTRS, &field_info_operator_call_rx },
	{ XFIELD_TYPE_FIELD, 2, 25, 10, NULL, FATTRS, &field_info_freq_rx },
	/* TODO: power */
	{ XFIELD_TYPE_FIELD, 2, 36,  6, NULL, FATTRS, &field_info_mode_rx },
	{ XFIELD_TYPE_FIELD, 2, 43,  3, NULL, FATTRS, &field_info_rst_rx },
	{ XFIELD_TYPE_FIELD, 2, 47, 16, NULL, FATTRS, &field_info_name_rx },
	{ XFIELD_TYPE_FIELD, 2, 64,  8, NULL, FATTRS, &field_info_grid_rx },
	{ XFIELD_TYPE_FIELD, 2, 73,  3, NULL, FATTRS, &field_info_zone_dxcc_rx },
	{ XFIELD_TYPE_FIELD, 2, 78,  2, NULL, FATTRS, &field_info_zone_itu_rx },
	{ XFIELD_TYPE_FIELD, 2, 82,  2, NULL, FATTRS, &field_info_zone_cqz_rx },
	{ XFIELD_TYPE_FIELD, 2, 86,  4, NULL, FATTRS, &field_info_seq_rx },

	/* row 3 */
	{ XFIELD_TYPE_LABEL, 3,  3, 35, &label_qth, LATTRS },
	{ XFIELD_TYPE_LABEL, 3, 39,  2, &label_state, LATTRS },
	{ XFIELD_TYPE_LABEL, 3, 42, 22, &label_county, LATTRS },
	{ XFIELD_TYPE_LABEL, 3, 65, 26, &label_country, LATTRS },
	{ XFIELD_TYPE_LABEL, 3, 92,  2, &label_continent, LATTRS },

	/* row 4 */
	{ XFIELD_TYPE_LABEL, 4,  0,  2, &label_de, LATTRS },
	{ XFIELD_TYPE_FIELD, 4,  3, 35, NULL, FATTRS, &field_info_qth_tx },
	{ XFIELD_TYPE_FIELD, 4, 39,  2, NULL, FATTRS, &field_info_state_tx },
	{ XFIELD_TYPE_FIELD, 4, 42, 22, NULL, FATTRS, &field_info_county_tx },
	{ XFIELD_TYPE_FIELD, 4, 65, 26, NULL, FATTRS, &field_info_country_tx },
	{ XFIELD_TYPE_FIELD, 4, 92,  2, NULL, FATTRS, &field_info_continent_tx },

	/* row 5 */
	{ XFIELD_TYPE_LABEL, 5,  0,  2, &label_dx, LATTRS },
	{ XFIELD_TYPE_FIELD, 5,  3, 35, NULL, FATTRS, &field_info_qth_rx },
	{ XFIELD_TYPE_FIELD, 5, 39,  2, NULL, FATTRS, &field_info_state_rx },
	{ XFIELD_TYPE_FIELD, 5, 42, 22, NULL, FATTRS, &field_info_county_rx },
	{ XFIELD_TYPE_FIELD, 5, 65, 26, NULL, FATTRS, &field_info_country_rx },
	{ XFIELD_TYPE_FIELD, 5, 92,  2, NULL, FATTRS, &field_info_continent_rx },
};

/* clean up just in case */
#undef LATTRS
#undef FATTRS

/* the contest fields */
static struct xfield_def *wfields;
static size_t nwfields;

static struct xform_palette palette = {
	.pair = {
		[XFORM_COLOR_BLACK] = COLOR_PAIR(COLOR_HINT_BLACK),
		[XFORM_COLOR_RED] = COLOR_PAIR(COLOR_HINT_RED),
		[XFORM_COLOR_GREEN] = COLOR_PAIR(COLOR_HINT_GREEN),
		[XFORM_COLOR_YELLOW] = COLOR_PAIR(COLOR_HINT_YELLOW),
		[XFORM_COLOR_BLUE] = COLOR_PAIR(COLOR_HINT_BLUE),
		[XFORM_COLOR_MAGENTA] = COLOR_PAIR(COLOR_HINT_MAGENTA),
		[XFORM_COLOR_CYAN] = COLOR_PAIR(COLOR_HINT_CYAN),
		[XFORM_COLOR_WHITE] = COLOR_PAIR(COLOR_HINT_WHITE),
	},
};

/*
 * contact setup
 */
int init_contact(lua_State *L, struct qso **qso, char *error, size_t errlen)
{
	struct qso *tmp;

	tmp = qso_alloc();
	if (!tmp)
		return -ENOMEM;

	qso_copy(tmp, template);

	/* switch to new entry */
	qso_putref(*qso);
	*qso = tmp;

	/* generate a new uuid - it's ugly that we have to reach in like this */
	xuuid_generate(&tmp->uuid);
	xuuid_unparse(&tmp->uuid, tmp->uuid_str);

	/* let the script finish up setup */
	return script_event_qso_init(L, tmp, error, errlen);
}

/*
 * location
 */
static void handle_location_notification(enum announce_loc_msg_type type,
					 int mode, const char *grid)
{
	MXLOCK(&location_lock);
	if (mode)
		strcpy_safe(location, grid, sizeof(location));
	else
		location[0] = '\0';
	MXUNLOCK(&location_lock);
}

static void draw_location(void)
{
	const char *grid;

	MXLOCK(&location_lock);
	if (location[0])
		grid = location;
	else
		grid = UNKNOWN_LOCATION_STR;

	mvwprintw(header_win, 0, 10, "%-*s", sizeof(location), grid);
	MXUNLOCK(&location_lock);
}

/*
 * rig
 */
static void handle_rig_notification(enum announce_rig_msg_type type,
				    uint64_t tx_freq, uint64_t rx_freq,
				    uint32_t power)
{
	MXLOCK(&rig_lock);
	rig_info.tx_freq = tx_freq;
	rig_info.rx_freq = rx_freq;
	rig_info.power = power;
	MXUNLOCK(&rig_lock);
}

static void __set_rig_info(struct qso_side *side, uint64_t v,
			   struct val *(*get)(const struct qso_side *),
			   int (*set)(struct qso_side *, struct val *),
			   void (*clr)(struct qso_side *))
{
	struct val *cur;

	if (!v)
		return;

	cur = get(side);
	if (!cur || (cur->type != VT_INT) || (cur->i != v)) {
		if (set(side, VAL_ALLOC_INT(v)))
			clr(side);
	}

	val_putref(cur);
}

#define __set_rig_info_freq(side, freq) \
	__set_rig_info((side), (freq), \
		       qso_get_freq, qso_set_freq_val, qso_clear_freq)
#define __set_rig_info_power(side, power) \
	__set_rig_info((side), (power), \
		       qso_get_power, qso_set_power_val, qso_clear_power)

static void set_rig_info(void)
{
	struct rig_info info;

	MXLOCK(&rig_lock);
	info = rig_info;
	MXUNLOCK(&rig_lock);

	__set_rig_info_freq(&contact->tx, info.tx_freq);
	__set_rig_info_freq(&contact->rx, info.rx_freq);
	__set_rig_info_power(&contact->tx, info.power);

	__set_rig_info_freq(&template->tx, info.tx_freq);
	__set_rig_info_freq(&template->rx, info.rx_freq);
	__set_rig_info_power(&template->tx, info.power);
}

/*
 * clock
 */
static void draw_clock(void)
{
	struct tm tm;
	time_t now;

	now = time(NULL);
	gmtime_r(&now, &tm);

	mvwprintw(header_win, 0, HEADER_WIN_WIDTH - 19, "%04u-%02u-%02u %02u:%02u:%02u",
		  tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
		  tm.tm_hour, tm.tm_min, tm.tm_sec);
}

/*
 * misc
 */
static void tweak_script_colors(void)
{
	size_t i;

	for (i = 0; i < nwfields; i++) {
		struct xfield_def *f = &wfields[i];

		/* labels should use color */
		if (f->type == XFIELD_TYPE_LABEL)
			f->attr.color = LABEL_XCOLOR;
	}
}

static void setup_windows(void)
{
	int ret;

	/*
	 * Draw the various separators/boxes/etc.
	 */
	/* left/right screen half separator for header */
	xwvline(stdscr, HEADER_WIN_ROW, HEADER_WIN_WIDTH,
		HEADER_WIN_HEIGHT, ACS_VLINE, ACS_VLINE, ACS_VLINE);

	/* header/supercheck+callout separator */
	whrule(stdscr, HEADER_WIN_HEIGHT, HEADER_WIN_WIDTH);

	/* 3 remaining sides around callout */
	xwvline(stdscr, CALLOUT_WIN_ROW - 1, CALLOUT_WIN_COL - 1,
		CALLOUT_WIN_HEIGHT + 2, ACS_TTEE, ACS_VLINE, ' ');
	xwvline(stdscr, CALLOUT_WIN_ROW - 1, CALLOUT_WIN_COL + CALLOUT_WIN_WIDTH,
		LINES - CALLOUT_WIN_ROW + 1, ACS_RTEE, ACS_VLINE, ACS_VLINE);
	xwhline(stdscr, CALLOUT_WIN_ROW + CALLOUT_WIN_HEIGHT,
		CALLOUT_WIN_COL - 1, CALLOUT_WIN_WIDTH + 2,
		ACS_LLCORNER, ACS_HLINE, ACS_RTEE);

	/* recent lists separator */
	xwhline(stdscr, RECENT_WIN_ROW - 1, 0, RECENT_WIN_WIDTH + 1,
		ACS_HLINE, ACS_HLINE, ACS_RTEE);

	wrefresh(stdscr); /* refresh all the lines */

	/* header with status */
	header_win = newwin(HEADER_WIN_HEIGHT, HEADER_WIN_WIDTH, HEADER_WIN_ROW,
			    HEADER_WIN_COL);
	mvwprintw(header_win, 0, 0, "Location:");
	mvwprintw(header_win, 0,
		  (HEADER_WIN_WIDTH - strlen(contest_params.name) - 1) / 2,
		  contest_params.name);
	draw_location();
	draw_clock();

	/* the callout */
	ret = callout_init(CALLOUT_WIN_HEIGHT, CALLOUT_WIN_WIDTH,
			   CALLOUT_WIN_ROW, CALLOUT_WIN_COL);
	ASSERT0(ret);

	/* partial callsign lookup results */
	supercheck_win = newwin(SUPERCHECK_WIN_HEIGHT, SUPERCHECK_WIN_WIDTH,
				SUPERCHECK_WIN_ROW, SUPERCHECK_WIN_COL);
	mvwprintw(supercheck_win, 0, 0, "Supercheck:");

	/* current QSO - contest exchange */
	ret = xform_alloc(&work, WORK_WIN_HEIGHT, WORK_WIN_WIDTH,
			  WORK_WIN_ROW, WORK_WIN_COL, &palette,
			  wfields, nwfields);
	ASSERT0(ret);

	/* status line - errors and QSO count */
	ret = status_init(STATUS_WIN_HEIGHT, STATUS_WIN_WIDTH,
			  STATUS_WIN_ROW, STATUS_WIN_COL,
			  ERROR_COLOR);
	ASSERT0(ret);

	/* recent QSOs for current call */
	ret = recent_init(&recent_work_list, RECENT_WORK_WIN_HEIGHT,
			  RECENT_WORK_WIN_WIDTH, RECENT_WORK_WIN_ROW,
			  RECENT_WORK_WIN_COL, false);
	recent_refresh(&recent_work_list, NULL);

	/* recent QSOs */
	ret = recent_init(&recent_list, RECENT_WIN_HEIGHT, RECENT_WIN_WIDTH,
			  RECENT_WIN_ROW, RECENT_WIN_COL, true);
	recent_refresh(&recent_list, NULL);

	/* current QSO - full */
	ret = xform_alloc(&current, CURRENT_WIN_HEIGHT, CURRENT_WIN_WIDTH,
			  CURRENT_WIN_ROW, CURRENT_WIN_COL, &palette,
			  cfields, ARRAY_LEN(cfields));
	ASSERT0(ret);
}

static void refresh_windows(void)
{
	wrefresh(header_win);
	wrefresh(supercheck_win);
	xform_refresh(&work);
	status_refresh();
	recent_refresh(&recent_work_list, qso_get_station_call(&contact->rx));
	recent_refresh(&recent_list, NULL);
	xform_refresh(&current);

	/* move cursor to the active form */
	xform_cursor_warp(active_form);
}

static void usage(const char *prog)
{
#define P(...) fprintf(stderr, __VA_ARGS__)

	P("HLog version %s (libjeffpc %s)\n\n", version_string, jeffpc_version);

	P("Usage: %s -c <contest> [-y <year>] [-E <path>] <data dir>\n", prog);
	P("\n");
	P("\t-E <path>    - export contest log to path\n");
	P("\t-c <contest> - name of contest (e.g., CQ-WW-SSB)\n");
	P("\t-y <year>    - year of contest (defaults to current year)\n");

	exit(1);

#undef P
}

static int load_index(void)
{
	int ret;

	printf("Loading existing QSOs...\n");
	ret = index_init(AT_FDCWD);
	printf("%s.\n", ret ? "error" : "done");

	if (ret)
		fprintf(stderr, "Error: failed to load existing entries: %s\n",
			xstrerror(ret));

	return ret;
}

static void sync_current(void)
{
	struct xform *form = &current;
	size_t i;

	if (active_form == &current)
		return; /* we're directly modifying the current form */

	for (i = 0; i < form->nfields; i++) {
		const struct field_info *info;
		struct val *val;

		info = xform_field_get_private(form, i);

		ASSERT3P(info, !=, NULL);

		switch (info->what) {
			case FI_SIDE_TX:
				val = info->get(&contact->tx);
				break;
			case FI_SIDE_RX:
				val = info->get(&contact->rx);
				break;
		}

		if (!val) {
			xform_field_clear(form, i);
			continue;
		}

		switch (val->type) {
			case VT_NULL:
				xform_field_clear(form, i);
				break;
			case VT_STR:
				xform_field_set(form, i, val_cstr(val));
				break;
			case VT_INT: {
				char tmp[30];

				snprintf(tmp, sizeof(tmp), "%"PRIu64, val->i);

				xform_field_set(form, i, tmp);
				break;
			}
			case VT_BOOL:
			case VT_CHAR:
			case VT_NVL:
			case VT_ARRAY:
			case VT_SYM:
			case VT_CONS:
			case VT_BLOB:
				panic("impossible");
		}

		val_putref(val);
	}
}

static int do_contest(lua_State *L)
{
	char error[128];
	int ret;

	tweak_script_colors();

	/* set up the current contact struct qso */
	ret = init_contact(L, &contact, error, sizeof(error));
	if (ret) {
		fprintf(stderr, "Error: Failed to initialize QSO: %s\n",
			error);
		return ret;
	}

	initscr();
	noecho();
	cbreak();
	halfdelay(5);
	start_color();

	init_pair(COLOR_HINT_BLACK, COLOR_BLACK, COLOR_BLACK); /* silly */
	init_pair(COLOR_HINT_RED, COLOR_RED, COLOR_BLACK);
	init_pair(COLOR_HINT_GREEN, COLOR_GREEN, COLOR_BLACK);
	init_pair(COLOR_HINT_YELLOW, COLOR_YELLOW, COLOR_BLACK);
	init_pair(COLOR_HINT_BLUE, COLOR_BLUE, COLOR_BLACK);
	init_pair(COLOR_HINT_MAGENTA, COLOR_MAGENTA, COLOR_BLACK);
	init_pair(COLOR_HINT_CYAN, COLOR_CYAN, COLOR_BLACK);
	init_pair(COLOR_HINT_WHITE, COLOR_WHITE, COLOR_BLACK);

	setup_windows();

	sync_current();
	refresh_windows();

	for (;;) {
		int ch;

		ch = xform_getch(active_form);

		set_rig_info(); /* do this here so lua calls below can use it */

		if (ch != ERR) {
			status_display_error(""); /* clear the error string */

			ui_work_dispatch(L, ch, &contact);
		}

		draw_location();
		draw_clock();

		recent_refresh(&recent_work_list, qso_get_station_call(&contact->rx));
		recent_refresh(&recent_list, NULL);
		sync_current();
		refresh_windows();
	}

	xform_free(&work);
	xform_free(&current);

	delwin(header_win);
	callout_destroy();
	delwin(supercheck_win);
	recent_destroy(&recent_work_list);
	recent_destroy(&recent_list);
	status_deinit();

	endwin();

	return 0;
}

static int do_export(lua_State *L, const char *outdir)
{
	char error[128];
	int ret;

	printf("Exporting %s entries to %s...\n", contest_params.name, outdir);

	ret = script_event_export(L, outdir, error, sizeof(error));
	if (ret)
		fprintf(stderr, "Error: Failed to export log: %s\n",
			error);

	return ret;
}

int main(int argc, char **argv)
{
	lua_State *L;
	char *export;
	int ret;
	int opt;

	setlocale(LC_ALL, "");
	ASSERT0(putenv("TZ=UTC"));

	export = NULL;
	while ((opt = getopt(argc, argv, "E:c:y:")) != -1) {
		switch (opt) {
			case 'E':
				export = realpath(optarg, NULL);
				if (!export) {
					fprintf(stderr, "Error: invalid export "
						"path specified: %s\n",
						strerror(errno));
					usage(argv[0]);
				}
				break;
			case 'c':
				contest_params.name = optarg;
				break;
			case 'y':
				ret = str2u32(optarg, &contest_params.year);
				if (ret || (contest_params.year < 1900) ||
				    (contest_params.year > 2200)) {
					fprintf(stderr, "Error: invalid year "
					       "'%s'\n\n", optarg);
					usage(argv[0]);
				}
				break;
			default:
				usage(argv[0]);
				break;
		}
	}

	if (!contest_params.name) {
		fprintf(stderr, "Error: contest name required\n\n");
		usage(argv[0]);
	}

	ret = script_exists(&contest_params);
	if (ret) {
		fprintf(stderr, "Error: unknown contest '%s'\n\n",
			contest_params.name);
		usage(argv[0]);
	}

	/* missing data dir path */
	if ((optind + 1) != argc) {
		fprintf(stderr, "Error: path to data directory required\n\n");
		usage(argv[0]);
	}

	MXINIT(&location_lock, &location_lc);
	MXINIT(&rig_lock, &rig_lc);

	data_dir = argv[argc - 1];

	/* contest year defaults to "now" */
	if (!contest_params.year) {
		struct tm tm;
		time_t now;

		now = time(NULL);
		gmtime_r(&now, &tm);

		contest_params.year = tm.tm_year + 1900;
	}

	/*
	 * chdir to the data directory to make our lives in the future much
	 * easier.  We will be able to use relative paths everywhere.
	 */
	if (chdir(data_dir)) {
		fprintf(stderr, "Error: failed to change to data dir: %s\n",
			strerror(errno));
		ret = -errno;
		goto err;
	}

	ret = load_databases();
	if (ret)
		goto err;

	ret = load_index();
	if (ret)
		goto err;

	if (!export) {
		ret = announce_connect(NULL, handle_location_notification,
				       handle_rig_notification);
		if (ret)
			goto err;

		template = qso_alloc();
		if (!template) {
			ret = -ENOMEM;
			goto err;
		}

		contact = qso_alloc();
		if (!contact) {
			ret = -ENOMEM;
			goto err_free;
		}

		ret = load_config(AT_FDCWD, template, "contest");
		if (ret)
			goto err_free;
	} else {
		template = NULL;
		contact = NULL;
	}

	L = script_load(&wfields, &nwfields, &contest_params, template);
	if (IS_ERR(L)) {
		ret = PTR_ERR(L);
		goto err_free;
	}

	if (!export)
		ret = do_contest(L);
	else
		ret = do_export(L, export);

	lua_close(L);

err_free:
	qso_putref(template);
	qso_putref(contact);

err:
	MXDESTROY(&rig_lock);
	MXDESTROY(&location_lock);

	free(export);

	return ret ? 1 : 0;
}
