/*
 * Copyright (c) 2020-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/version.h>
#include <jeffpc/error.h>
#include <jeffpc/time.h>
#include <jeffpc/val.h>
#include <jeffpc/io.h>
#include <jeffpc/unicode.h>
#include <jeffpc/cstr.h>

#include <ctype.h>
#include <locale.h>
#include <math.h>

#include <hlog/version.h>
#include <hlog/qso.h>
#include <hlog/qso_line.h>
#include <hlog/adif.h>
#include <hlog/bands.h>
#include <hlog/index.h>
#include <hlog/country.h>
#include <hlog/maidenhead.h>
#include <hlog/url.h>
#include <hlog/util.h>
#include <hlog/zones.h>
#include <hlog-lua/xlua.h>

#include "common/config.h"
#include "common/databases.h"
#include "common/recent.h"
#include "common/status.h"
#include "common/ui.h"
#include "common/xform.h"
#include "common/xcurses.h"

#define EARTH_RADIUS		6378.137

#define HEADER_WIN_HEIGHT	1
#define HEADER_WIN_WIDTH	80
#define HEADER_WIN_ROW		0
#define ENTRY_WIN_HEIGHT	6
#define ENTRY_WIN_WIDTH		80
#define ENTRY_WIN_ROW		1
#define CALL_WIN_HEIGHT		10
#define CALL_WIN_WIDTH		ENTRY_WIN_WIDTH
#define CALL_WIN_ROW		(ENTRY_WIN_ROW + ENTRY_WIN_HEIGHT)
#define CMD_WIN_HEIGHT		1
#define CMD_WIN_WIDTH		RECENT_WIN_WIDTH
#define CMD_WIN_ROW		(LINES - 2)
#define STATUS_WIN_HEIGHT	1
#define STATUS_WIN_WIDTH	RECENT_WIN_WIDTH
#define STATUS_WIN_ROW		(LINES - 1)
#define RECENT_WIN_HEIGHT	(LINES - ENTRY_WIN_HEIGHT - CALL_WIN_HEIGHT - \
				 CMD_WIN_HEIGHT - STATUS_WIN_HEIGHT)
#define RECENT_WIN_WIDTH	CALL_WIN_WIDTH
#define RECENT_WIN_ROW		(CALL_WIN_ROW + CALL_WIN_HEIGHT)

#define CMD_HINT_COLOR		1

#define MAX_STATE_COUNT		10 /* number of concurrent states (tabs) */
#define MAX_SUMMARY_MODE_COUNT	4 /* number of modes to display in call summary */
#define MAX_SUMMARY_BAND_COUNT	15 /* number of bands to display in call summary */

static const char *data_dir;

static struct str default_date = STR_STATIC_INITIALIZER("????""-??""-??");
static struct str default_time = STR_STATIC_INITIALIZER("??:??:??");
static struct str default_latlon = STR_STATIC_INITIALIZER("??.??N ???.??E");
static struct str default_azidist =
	STR_STATIC_INITIALIZER("???° ????? km/???° ????? km");

/* various parts of the display */
enum refresh {
	REF_DE =		0x00000001,
	REF_DE_GRID =		0x00000002,
	REF_DE_POWER =		0x00000004,
	REF_DATE =		0x00000008,
	REF_TIME_START =	0x00000010,
	REF_TIME_END =		0x00000020,
	REF_CALL =		0x00000040,
	REF_FREQ =		0x00000080,
	REF_MODE =		0x00000100,
	REF_RSTS =		0x00000200,
	REF_RSTR =		0x00000400,
	REF_NAME =		0x00000800,
	REF_QTH =		0x00001000,
	REF_GRID =		0x00002000,
	REF_COUNTRY =		0x00004000,
	REF_GEO_LATLON =	0x00008000,
	REF_GEO_AZIDIST =	0x00010000,
	REF_STATE =		0x00020000,
	REF_COUNTY =		0x00040000,
	REF_ZONE_CONT =		0x00080000,
	REF_ZONE_DXCC =		0x00100000,
	REF_ZONE_ITU =		0x00200000,
	REF_ZONE_CQZ =		0x00400000,
#define REF_ZONES	(REF_ZONE_CONT | REF_ZONE_DXCC | REF_ZONE_ITU | \
			 REF_ZONE_CQZ)
	REF_COMMENT =		0x00800000,
	REF_CALL_WIN =		0x40000000,
	REF_RECENT_WIN =	0x80000000,
};

/* This contains everything being displayed on screen */
struct state {
	struct qso *qso;

	/* the following are generated from the QSO data above */
	struct str *date;
	struct str *time_start;
	struct str *time_end;
	struct str *latlon; /* lat & lon */
	struct str *azidist; /* azimuth & distance */

	enum refresh dirty; /* which fields have been modified since last save */
	bool in_use; /* is this state in-use */
};

struct stats_summary {
	struct {
		struct str *mode_name;
		uint64_t band[MAX_SUMMARY_BAND_COUNT];
		uint64_t total; /* number of time this mode has been used */
	} mode[MAX_SUMMARY_MODE_COUNT];
};

/* list of bands (in mm) in call summary table */
static uint64_t summary_bands[MAX_SUMMARY_BAND_COUNT] = {
	0,
	160000,
	80000,
	60000,
	40000,
	30000,
	20000,
	17000,
	15000,
	12000,
	10000,
	6000,
	2000,
	700,
	230,
};

static struct qso *template;
static struct state *current;
static struct state states[MAX_STATE_COUNT];

struct entry_field {
	int row;
	int col_label;
	int col_value;
	/*
	 * The field label.  All characters other than '_' are printed.  A
	 * '_' indicates that the following character should be highlighted
	 * to hint at the command to modify this field.
	 */
	const char *label;
	size_t len;
	enum refresh refresh;
	bool tx_side; /* only meaningful for get.side */
	struct {
		struct val *(*state)(void);
		struct val *(*qso)(const struct qso *);
		struct val *(*side)(const struct qso_side *);
	} get;
};

enum cmd_result {
	CR_OK = 0,		/* success */
	CR_TOO_MANY_TABS,	/* cannot open new tab */
	CR_INVAL_TAB_ID,	/* malformed tab id */
	CR_UNKNOWN_TAB_ID,	/* unknown tab id */
};

struct command {
	const char *cmd;
	enum cmd_result (*fxn)(char *arg, char *error, size_t errorlen);
	bool arg_required;
	enum refresh refresh;
};

static enum cmd_result cmd_tab(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_new(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_reset(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_save(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_edit(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_start(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_end(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_startend(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_qrz(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_ae7q(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_de(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_de_grid(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_de_power(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_power(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_call(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_freq(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_mode(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_rst_s(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_rst_r(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_name(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_qth(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_grid(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_country(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_state(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_county(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_continent(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_dxcc(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_itu(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_cqz(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_comment(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_seq_r(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_seq_s(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_pota(char *arg, char *error, size_t errorlen);
static enum cmd_result cmd_wwff(char *arg, char *error, size_t errorlen);

static struct val *state_get_date(void);
static struct val *state_get_start_time(void);
static struct val *state_get_end_time(void);
static struct val *state_get_latlon(void);
static struct val *state_get_azidist(void);

static const struct entry_field entry_fields[] = {
	{ 0,  0,  3, "_d_e",          10, REF_DE, true,
	  { .side = qso_get_station_call_val }, },
	{ 0, 15, 21, "_Grid:",         8, REF_DE_GRID, true,
	  { .side = qso_get_grid_val }, },
	{ 0, 31, 36, "_P_w_r:",        5, REF_DE_POWER, true,
	  { .side = qso_get_power }, },
	{ 0,  0, 52, NULL,            10, REF_DATE, false,
	  { .state = state_get_date }, },
	{ 0,  0, 63, NULL,             8, REF_TIME_START, false,
	  { .state = state_get_start_time }, },
	{ 0,  0, 72, NULL,             8, REF_TIME_END, false,
	  { .state = state_get_end_time }, },
	{ 1,  0,  6, "_Call:",        10, REF_CALL, false,
	  { .side = qso_get_station_call_val }, },
	{ 1, 18, 24, "_Freq:",        10, REF_FREQ, true,
	  { .side = qso_get_freq }, },
	{ 1, 36, 42, "_Mode:",         6, REF_MODE, true,
	  { .side = qso_get_mode_val }, },
	{ 1, 50, 56, "_RST_s:",        3, REF_RSTS, true,
	  { .side = qso_get_rst_val }, },
	{ 1, 61, 67, "_RST_r:",        3, REF_RSTR, false,
	  { .side = qso_get_rst_val }, },
	{ 2,  0,  6, "_Na_me:",       16, REF_NAME, false,
	  { .side = qso_get_name_val }, },
	{ 2, 24, 29, "_Q_T_H:",       35, REF_QTH, false,
	  { .side = qso_get_qth_val }, },
	{ 2, 66, 72, "_Grid:",         8, REF_GRID, false,
	  { .side = qso_get_grid_val }, },
	{ 3,  0,  9, "_C_o_u_n_t_r_y:", 26, REF_COUNTRY, false,
	  { .side = qso_get_country_val }, },
	{ 3,  0, 37, NULL,            14, REF_GEO_LATLON, false,
	  { .state = state_get_latlon }, },
	{ 3,  0, 53, NULL,            27, REF_GEO_AZIDIST, false,
	  { .state = state_get_azidist }, },
	{ 4,  0,  7, "_S_t/Pr:",       2, REF_STATE, false,
	  { .side = qso_get_state_val }, },
	{ 4, 11, 19, "_C_o_u_n_t_y:", 22, REF_COUNTY, false,
	  { .side = qso_get_county_val }, },
	{ 4, 43, 49, "_C_o_n_t:",      2, REF_ZONE_CONT, false,
	  { .side = qso_get_continent_val }, },
	{ 4, 53, 59, "_D_X_C_C:",      3, REF_ZONE_DXCC, false,
	  { .side = qso_get_zone_dxcc }, },
	{ 4, 64, 69, "_I_T_U:",        2, REF_ZONE_ITU, false,
	  { .side = qso_get_zone_itu }, },
	{ 4, 73, 78, "_C_Q_Z:",        2, REF_ZONE_CQZ, false,
	  { .side = qso_get_zone_cqz }, },
	{ 5,  0,  9, "_Co_mment:",    71, REF_COMMENT, false,
	  { .qso = qso_get_comment_val }, },
};

static const struct command commands[] = {
	{ "tab",    cmd_tab,		true,  ~0u },
	{ "new",    cmd_new,		false, ~0u },
	{ "reset",  cmd_reset,		false, ~0u },
	{ "save",   cmd_save,		false, ~0u /* clears dirty bits */ },
	{ "edit",   cmd_edit,		true,  ~0u },
	{ "start",  cmd_start,		true,  REF_DATE | REF_TIME_START, },
	{ "end",    cmd_end,		true,  REF_DATE | REF_TIME_END, },
	{ "startend", cmd_startend,	true,  REF_DATE | REF_TIME_START |
					       REF_TIME_END, },
	{ "qrz",    cmd_qrz,		false, 0 },
	{ "ae7q",   cmd_ae7q,		false, 0 },

	{ "de",     cmd_de,		true,  REF_DE, },
	{ "dg",     cmd_de_grid,	true,  REF_DE_GRID | REF_GEO_AZIDIST, },
	{ "dpwr",   cmd_de_power,	true,  REF_DE_POWER, },
	{ "pwr",    cmd_power,		true,  0, },
	{ "c",      cmd_call,		true,  REF_CALL | REF_COUNTRY |
					       REF_GEO_LATLON | REF_GEO_AZIDIST |
					       REF_ZONES | REF_CALL_WIN, },
	{ "f",      cmd_freq,		true,  REF_FREQ, },
	{ "m",      cmd_mode,		true,  REF_MODE, },
	{ "rs",     cmd_rst_s,		true,  REF_RSTS, },
	{ "rr",     cmd_rst_r,		true,  REF_RSTR, },
	{ "nm",     cmd_name,		true,  REF_NAME, },
	{ "qth",    cmd_qth,		true,  REF_QTH, },
	{ "g",      cmd_grid,		true,  REF_GRID | REF_GEO_LATLON |
					       REF_GEO_AZIDIST, },
	{ "country",cmd_country,	true,  REF_COUNTRY, },
	{ "st",     cmd_state,		true,  REF_STATE | REF_ZONES, },
	{ "county", cmd_county,		true,  REF_COUNTY, },
	{ "cont",   cmd_continent,	true,  REF_ZONE_CONT, },
	{ "dxcc",   cmd_dxcc,		true,  REF_ZONES, },
	{ "itu",    cmd_itu,		true,  REF_ZONE_ITU, },
	{ "cqz",    cmd_cqz,		true,  REF_ZONE_CQZ, },
	{ "cm",     cmd_comment,	true,  REF_COMMENT, },
	{ "seqr",   cmd_seq_r,		true,  0 },
	{ "seqs",   cmd_seq_s,		true,  0 },
	{ "pota",   cmd_pota,		true,  0 },
	{ "wwff",   cmd_wwff,		true,  0 },
};

static WINDOW *header_win;
static WINDOW *entry_win;
static WINDOW *call_win;
static WINDOW *recent_win;

static struct xform cmdform;

static struct str label_cmd_prompt = STR_STATIC_INITIALIZER(">");

static const struct xfield_def cmdfields[] = {
	{ XFIELD_TYPE_LABEL, 0, 0, 1, &label_cmd_prompt,
	  { XFORM_COLOR_NONE, XFORM_UNDERLINE_AUTO }, NULL },
	{ XFIELD_TYPE_FIELD, 0, 2, CMD_WIN_WIDTH - 2, NULL,
	  { XFORM_COLOR_NONE, XFORM_UNDERLINE_OFF }, NULL},
};

/* print width amount of chars of string, padding or truncating as necessary */
static inline void wprintw_cstr(WINDOW *win, const char *str, size_t width)
{
	size_t inlen;

	if (!str)
		str = "";

	inlen = strlen(str);

	if (inlen >= width)
		wprintw(win, "%-*.*s ", width, width, str);
	else
		wprintw(win, "%-*s ", width, str);
}

/*
 * State refresh helpers
 */
static struct val *state_get_date(void)
{
	return str_getref_val(current->date);
}

static struct val *state_get_start_time(void)
{
	return str_getref_val(current->time_start);
}

static struct val *state_get_end_time(void)
{
	return str_getref_val(current->time_end);
}

static struct val *state_get_latlon(void)
{
	return str_getref_val(current->latlon);
}

static struct val *state_get_azidist(void)
{
	return str_getref_val(current->azidist);
}

/*
 * QSO reseting logic
 *
 * This is custom logic to partially reset the QSO structure.  This allows
 * us to preserve some of the fields (e.g., frequency or mode) between QSOs.
 */
static void reset_qso(struct qso *qso)
{
	qso_clear_datetime(qso, true);
	qso_clear_datetime(qso, false);
	qso_clear_comment(qso);

	/* only the RST and seq change on the tx side */
	qso_clear_rst(&qso->tx);
	qso_clear_seq(&qso->tx);

	/* everything but the frequency and mode changes on the rx side */
	qso_clear_station_call(&qso->rx);
	qso_clear_operator_call(&qso->rx);
	qso_clear_power(&qso->rx);
	qso_clear_rst(&qso->rx);
	qso_clear_name(&qso->rx);
	qso_clear_qth(&qso->rx);
	qso_clear_grid(&qso->rx);
	qso_clear_country(&qso->rx);
	qso_clear_state(&qso->rx);
	qso_clear_county(&qso->rx);
	qso_clear_continent(&qso->rx);
	qso_clear_zone_dxcc(&qso->rx);
	qso_clear_zone_itu(&qso->rx);
	qso_clear_zone_cqz(&qso->rx);
	qso_clear_seq(&qso->rx);

	nvl_putref(qso->additional);
	qso->additional = nvl_alloc();
	ASSERT(!IS_ERR(qso->additional));

	xuuid_generate(&qso->uuid);
	xuuid_unparse(&qso->uuid, qso->uuid_str);
}

/*
 * Command handler helpers
 */
/* returns NULL on error */
static struct state *tab_alloc(void)
{
	int tab;

	for (tab = 0; tab < ARRAY_LEN(states); tab++) {
		if (states[tab].in_use)
			continue;

		states[tab].in_use = true;
		return &states[tab];
	}

	return NULL;
}

static void tab_close(struct state *tab)
{
	tab->in_use = false;
}

/* returns NULL on error */
static struct state *tab_find_first_inuse(void)
{
	int tab;

	for (tab = 0; tab < ARRAY_LEN(states); tab++) {
		if (!states[tab].in_use)
			continue;

		return &states[tab];
	}

	return NULL;
}

/* returns negated errno on error */
static struct state *tab_find_id(const char *str)
{
	uint32_t tab;
	int ret;

	ret = str2u32(str, &tab);
	if (ret)
		return ERR_PTR(ret);

	if (tab >= MAX_STATE_COUNT)
		return ERR_PTR(-EINVAL);

	return &states[tab];
}

static void __activate_tab(struct state *tab)
{
	ASSERT(tab->in_use);
	current = tab;
}

static void __recalculate_time(const struct time *time, struct str **out)
{
	char buf[9];

	if (time_snprint_time(buf, sizeof(buf), time))
		*out = STR_DUP(buf);
	else
		*out = &default_time;
}

static void recalculate_times(void)
{
	struct time *date;
	char buf[11];

	str_putref(current->date);
	str_putref(current->time_start);
	str_putref(current->time_end);

	if (current->qso->start.have & TIME_PART_DATE)
		date = &current->qso->start;
	else if (current->qso->end.have & TIME_PART_DATE)
		date = &current->qso->end;
	else
		date = NULL;

	if (time_snprint_date(buf, sizeof(buf), date))
		current->date = STR_DUP(buf);
	else
		current->date = &default_date;

	__recalculate_time(&current->qso->start, &current->time_start);
	__recalculate_time(&current->qso->end, &current->time_end);
}

static void recalculate_location(void)
{
	struct str *grid;
	struct str *call;
	double lat, lon;

	str_putref(current->latlon);

	grid = qso_get_grid(&current->qso->rx);
	call = qso_get_station_call(&current->qso->rx);

	if (maidenhead_decode(str_cstr(grid), &lat, &lon) ||
	    country_latlon(str_cstr(call), &lat, &lon)) {
		char buf[15];
		char ns, ew;

		ns = (lat >= 0) ? 'N' : 'S';
		ew = (lon >= 0) ? 'E' : 'W';

		snprintf(buf, sizeof(buf), "%5.2f%c %6.2f%c",
			 fabs(lat), ns,
			 fabs(lon), ew);

		current->latlon = STR_DUP(buf);
	} else {
		current->latlon = &default_latlon;
	}

	str_putref(grid);
	str_putref(call);
}

static void recalculate_distance(void)
{
	struct qso *qso = current->qso;
	struct str *de_grid, *dx_grid;
	struct str *dx_call;
	double de_lat, de_lon;
	double dx_lat, dx_lon;
	double distance;
	double azimuth;
	unsigned int int_azi;
	char buf[30];

	str_putref(current->azidist);

	de_grid = qso_get_grid(&qso->tx);

	if (!de_grid)
		goto not_calculated; /* no idea where we are */
	if (!maidenhead_decode(str_cstr(de_grid), &de_lat, &de_lon)) {
		str_putref(de_grid);
		goto not_calculated; /* no idea where we are */
	}

	dx_grid = qso_get_grid(&qso->rx);
	dx_call = qso_get_station_call(&qso->rx);

	if (!dx_call && !dx_grid)
		goto not_calculated; /* no idea where they are */
	if (!maidenhead_decode(str_cstr(dx_grid), &dx_lat, &dx_lon) &&
	    !country_latlon(str_cstr(dx_call), &dx_lat, &dx_lon)) {
		str_putref(dx_grid);
		str_putref(dx_call);
		goto not_calculated; /* no idea where they are */
	}

	/* degrees to radians */
	de_lat *= M_PI / 180.;
	de_lon *= M_PI / 180.;
	dx_lat *= M_PI / 180.;
	dx_lon *= M_PI / 180.;

	/*
	 * The following assumes a spherical earth.  This makes all the
	 * numbers slightly off from what we'd get if we used WGS84
	 * math...which we should.
	 */

	/* calculate */
	distance = acos(sin(de_lat) * sin(dx_lat) +
			cos(de_lat) * cos(dx_lat) * cos(de_lon - dx_lon));
	azimuth = fmod(atan2(sin(dx_lon - de_lon) * cos(dx_lat),
			     cos(de_lat) * sin(dx_lat) -
			     sin(de_lat) * cos(dx_lat) * cos(dx_lon - de_lon)),
		       2 * M_PI);

	/* radians to km & degrees */
	distance *= (2 * M_PI * EARTH_RADIUS) / (2 * M_PI);
	azimuth *= 180 / M_PI;

	/* convert to an int */
	int_azi = (azimuth >= 0.) ? azimuth : (azimuth + 360.);

	ASSERT3U(int_azi, <, 360);

	snprintf(buf, sizeof(buf), "%3u° %5.0f km/%3u° %5.0f km",
		 int_azi, distance,
		 (int_azi < 180) ? (int_azi + 180) : (int_azi - 180),
		 2 * M_PI * EARTH_RADIUS - distance);

	current->azidist = STR_DUP(buf);
	return;

not_calculated:
	current->azidist = &default_azidist;
}

static void refresh_zones(void)
{
	struct str *state;
	struct val *dxcc;
	int ret;

	dxcc = qso_get_zone_dxcc(&current->qso->rx);
	state = qso_get_state(&current->qso->rx);

	if (dxcc && state) {
		/* we have a country (dxcc) and state */
		ASSERT3U(dxcc->type, ==, VT_INT);

		ret = cqz_by_state(dxcc->i, str_cstr(state));
		if (ret >= 0) {
			qso_set_zone_cqz(&current->qso->rx, ret);

			current->dirty |= REF_ZONE_CQZ;
		}

		ret = ituz_by_state(dxcc->i, str_cstr(state));
		if (ret >= 0) {
			qso_set_zone_itu(&current->qso->rx, ret);

			current->dirty |= REF_ZONE_ITU;
		}
	}

	val_putref(dxcc);
	str_putref(state);
}

/*
 * Command handlers
 */
static enum cmd_result cmd_tab(char *arg, char *error, size_t errorlen)
{
	struct state *new;

	if (!strcasecmp(arg, "new")) {
		/* create a new tab */
		new = tab_alloc();
		if (!new)
			return CR_TOO_MANY_TABS;

		__activate_tab(new);
	} else if (!strcasecmp(arg, "close")) {
		/* close an existing tab */
		tab_close(current);

		new = tab_find_first_inuse();
		if (new) {
			__activate_tab(new);
			return CR_OK;
		}

		/* all tabs closed, open a new one */

		new = tab_alloc(); /* cannot fail, all tabs closed */
		ASSERT3P(new, !=, NULL);

		__activate_tab(new);
	} else {
		/* switch to an existing tab */
		new = tab_find_id(arg);
		if (IS_ERR(new))
			return CR_INVAL_TAB_ID;
		if (!new->in_use)
			return CR_UNKNOWN_TAB_ID;

		__activate_tab(new);
		return CR_OK;
	}

	/* reset all the on-screen state to start from scratch */
	reset_qso(current->qso);
	recalculate_times();
	recalculate_location();
	recalculate_distance();

	/* nothing is dirty - we reset it all */
	current->dirty = 0;

	return CR_OK;
}

static enum cmd_result cmd_new(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_save(current->qso, AT_FDCWD);
	if (ret) {
		wprintw(recent_win, "Failed to save QSO: %s",
			xstrerror(ret));
		return ret; /* returning a negated errno */
	}

	ret = index_save(current->qso);
	if (ret)
		wprintw(recent_win, "Warning: Failed to add QSO to index, "
			"UI may be out-of-sync: %s", xstrerror(ret));
	/* ignoring the index failure */

	/* display the uuid in the info message */
	snprintf(error, errorlen, "Saved %s", current->qso->uuid_str);

	reset_qso(current->qso);

	recalculate_times();
	recalculate_location();
	recalculate_distance();

	current->dirty = 0;

	return CR_OK;
}

static enum cmd_result cmd_reset(char *arg, char *error, size_t errorlen)
{
	reset_qso(current->qso);

	recalculate_times();
	recalculate_location();
	recalculate_distance();

	current->dirty = 0;

	return CR_OK;
}

static enum cmd_result cmd_save(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_save(current->qso, AT_FDCWD);
	if (ret) {
		wprintw(recent_win, "Failed to save QSO: %s",
			xstrerror(ret));
		return ret; /* returning a negated errno */
	}

	ret = index_save(current->qso);
	if (ret)
		wprintw(recent_win, "Warning: Failed to add QSO to index, "
			"UI may be out-of-sync: %s", xstrerror(ret));
	/* ignoring the index failure */

	/* display the uuid in the info message */
	snprintf(error, errorlen, "Saved %s", current->qso->uuid_str);

	current->dirty = 0;

	return CR_OK;
}

static enum cmd_result cmd_edit(char *arg, char *error, size_t errorlen)
{
	struct state *tab, *prev;
	struct xuuid uuid;
	int ret;

	if (!xuuid_parse(&uuid, arg))
		return -EINVAL;

	prev = current;

	/* open & activate a new tab */
	tab = tab_alloc();
	if (!tab)
		return CR_TOO_MANY_TABS;

	__activate_tab(tab);

	/* load entry from index */
	ret = index_get(&uuid, current->qso);
	if (ret) {
		tab_close(current);
		__activate_tab(prev);
		return ret;
	}

	/* calculate derived values */
	recalculate_times();
	recalculate_location();
	recalculate_distance();

	/* everything is clean - we just loaded it */
	current->dirty = 0;

	return CR_OK;
}

static enum cmd_result __cmd_startend(char *arg, char *error, size_t errorlen,
				      bool start, enum refresh dirty)
{
	int ret;

	if (!strcasecmp(arg, "now"))
		ret = qso_set_datetime(current->qso, start, NULL);
	else
		ret = qso_set_datetime(current->qso, start, arg);

	if (ret) {
		snprintf(error, errorlen, "invalid date or time, specify "
			 "'now', a date, or a time");
		return ret;
	}

	recalculate_times();

	current->dirty |= dirty;

	return CR_OK;
}

static enum cmd_result cmd_start(char *arg, char *error, size_t errorlen)
{
	return __cmd_startend(arg, error, errorlen, true,
			      REF_DATE | REF_TIME_START);
}

static enum cmd_result cmd_end(char *arg, char *error, size_t errorlen)
{
	return __cmd_startend(arg, error, errorlen, false,
			      REF_DATE | REF_TIME_END);
}

static enum cmd_result cmd_startend(char *arg, char *error, size_t errorlen)
{
	enum cmd_result res;

	res = __cmd_startend(arg, error, errorlen, true,
			     REF_DATE | REF_TIME_START);
	if (res != CR_OK)
		return res;

	return __cmd_startend(arg, error, errorlen, false,
			      REF_DATE | REF_TIME_END);
}

static enum cmd_result cmd_qrz(char *arg, char *error, size_t errorlen)
{
	struct str *call;
	char url[256];
	int ret;

	call = qso_get_station_call(&current->qso->rx);

	if (!call) {
		snprintf(error, errorlen, "No callsign entered");
		return -EINVAL;
	}

	snprintf(url, sizeof(url), "https://qrz.com/db/%s", str_cstr(call));

	ret = url_open(url);
	if (ret)
		snprintf(error, errorlen, "Failed to open browser: %s",
			 xstrerror(ret));
	else
		snprintf(error, errorlen, "Opening qrz.com...");

	str_putref(call);

	return ret;
}

static enum cmd_result cmd_ae7q(char *arg, char *error, size_t errorlen)
{
	struct str *call;
	char url[256];
	int ret;

	call = qso_get_station_call(&current->qso->rx);

	if (!call) {
		snprintf(error, errorlen, "No callsign entered");
		return -EINVAL;
	}

	snprintf(url, sizeof(url), "http://www.ae7q.com/query/data/CallHistory.php?CALL=%s",
		 str_cstr(call));

	ret = url_open(url);
	if (ret)
		snprintf(error, errorlen, "Failed to open browser: %s",
			 xstrerror(ret));
	else
		snprintf(error, errorlen, "Opening ae7q.com...");

	str_putref(call);

	return ret;
}

static enum cmd_result cmd_de(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_station_call(&current->qso->tx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set callsign: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_DE;

	return CR_OK;
}

static enum cmd_result cmd_de_grid(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_grid(&current->qso->tx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set grid: %s",
			 xstrerror(ret));
		return ret;
	}

	recalculate_distance();

	current->dirty |= REF_DE_GRID;

	return CR_OK;
}

static enum cmd_result cmd_de_power(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_power_user(&current->qso->tx, arg);
	if (ret) {
		snprintf(error, errorlen, "invalid TX power, specify "
			 "power in W without units (e.g., 6.7)");
		return ret;
	}

	current->dirty |= REF_DE_POWER;

	return CR_OK;
}

static enum cmd_result cmd_power(char *arg, char *error, size_t errorlen)
{
	char power[32];
	int ret;

	ret = qso_set_power_user(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "invalid TX power, specify "
			 "power in W without units (e.g., 6.7)");
		return ret;
	}

	pretty_print_power(power, sizeof(power),
			   qso_get_power(&current->qso->rx));

	snprintf(error, errorlen, "Set rx power to: %s", power);

	return CR_OK;
}

static enum cmd_result cmd_call(char *arg, char *error, size_t errorlen)
{
	struct country_result result;
	struct str *call;
	int ret;

	ret = qso_set_station_call(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set callsign: %s",
			 xstrerror(ret));
		return ret;
	}

	call = qso_get_station_call(&current->qso->rx);
	ASSERT3P(call, !=, NULL);

	if (country_lookup(str_cstr(call), &result)) {
		qso_set_country_str(&current->qso->rx, result.fullname);
		qso_set_continent(&current->qso->rx,
				  str_cstr(result.continent));
		qso_set_zone_dxcc(&current->qso->rx, result.code);

		str_putref(result.continent);

		/* dxcc change -> possiblity of CQ/ITU zone change */
		refresh_zones();
	} else {
		qso_clear_country(&current->qso->rx);
		qso_clear_continent(&current->qso->rx);
		qso_clear_zone_dxcc(&current->qso->rx);
		qso_clear_zone_itu(&current->qso->rx);
		qso_clear_zone_cqz(&current->qso->rx);
	}

	str_putref(call);

	recalculate_location();
	recalculate_distance();

	current->dirty |= REF_CALL;

	return CR_OK;
}

static enum cmd_result cmd_freq(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_freq(&current->qso->tx, arg);
	if (ret) {
		snprintf(error, errorlen, "invalid frequency, specify "
			 "frequency in MHz without units (e.g., 14.230)");
		return ret;
	}

	ret = qso_set_freq(&current->qso->rx, arg);
	ASSERT0(ret); /* it worked on the tx side */

	current->dirty |= REF_FREQ;

	return CR_OK;
}

static enum cmd_result cmd_mode(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_mode(&current->qso->tx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set transmit mode: %s",
			 xstrerror(ret));
		return ret;
	}

	ret = qso_set_mode(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set receive mode: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_MODE;

	return CR_OK;
}

static enum cmd_result cmd_rst_s(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_rst(&current->qso->tx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set sent RST: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_RSTS;

	return CR_OK;
}

static enum cmd_result cmd_rst_r(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_rst(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set received RST: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_RSTR;

	return CR_OK;
}

static enum cmd_result cmd_name(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_name(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set name: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_NAME;

	return CR_OK;
}

static enum cmd_result cmd_qth(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_qth(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set QTH: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_QTH;

	return CR_OK;
}

static enum cmd_result cmd_grid(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_grid(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set grid: %s",
			 xstrerror(ret));
		return ret;
	}

	recalculate_location();
	recalculate_distance();

	current->dirty |= REF_GRID;

	return CR_OK;
}

static enum cmd_result cmd_country(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_country(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set country: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_COUNTRY;

	return CR_OK;
}

static enum cmd_result cmd_state(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_state(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set state: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_STATE;

	/* state change -> possibility of CQ/ITU zone change */
	refresh_zones();

	return CR_OK;
}

static enum cmd_result cmd_county(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_county(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set county: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_COUNTY;

	return CR_OK;
}

static enum cmd_result cmd_continent(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_continent(&current->qso->rx, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set continent: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_ZONE_CONT;

	return CR_OK;
}

static enum cmd_result cmd_dxcc(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_zone_dxcc_user(&current->qso->rx, arg);
	switch (ret) {
		case 0:
			break;
		case -EINVAL:
		case -ERANGE:
			snprintf(error, errorlen, "invalid DXCC entity number");
			return ret;
		default:
			return ret;
	}

	current->dirty |= REF_ZONE_DXCC;

	/* dxcc change -> possibility of CQ/ITU zone change */
	refresh_zones();

	return CR_OK;
}

static enum cmd_result cmd_itu(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_zone_itu_user(&current->qso->rx, arg);
	switch (ret) {
		case 0:
			break;
		case -EINVAL:
		case -ERANGE:
			snprintf(error, errorlen, "invalid ITU zone, valid "
				 "range is 1..90, 0 clears the field");
			return ret;
		default:
			return ret;
	}

	current->dirty |= REF_ZONE_ITU;

	return CR_OK;
}

static enum cmd_result cmd_cqz(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_zone_cqz_user(&current->qso->rx, arg);
	switch (ret) {
		case 0:
			break;
		case -EINVAL:
		case -ERANGE:
			snprintf(error, errorlen, "invalid CQ zone, valid "
				 "range is 1..40, 0 clears the field");
			return ret;
		default:
			return ret;
	}

	current->dirty |= REF_ZONE_CQZ;

	return CR_OK;
}

static enum cmd_result cmd_comment(char *arg, char *error, size_t errorlen)
{
	int ret;

	ret = qso_set_comment(current->qso, arg);
	if (ret) {
		snprintf(error, errorlen, "Failed to set comment: %s",
			 xstrerror(ret));
		return ret;
	}

	current->dirty |= REF_COMMENT;

	return CR_OK;
}

static enum cmd_result __cmd_seq(char *arg, char *error, size_t errorlen,
				 struct qso_side *side)
{
	int ret;

	ret = qso_set_seq_user(side, arg);
	switch (ret) {
		case 0:
			break;
		case -EINVAL:
		case -ERANGE:
			snprintf(error, errorlen, "invalid serial: %s",
				 xstrerror(ret));
			return ret;
		default:
			return ret;
	}

	return CR_OK;
}

static enum cmd_result cmd_seq_r(char *arg, char *error, size_t errorlen)
{
	return __cmd_seq(arg, error, errorlen, &current->qso->rx);
}

static enum cmd_result cmd_seq_s(char *arg, char *error, size_t errorlen)
{
	return __cmd_seq(arg, error, errorlen, &current->qso->tx);
}

static enum cmd_result __cmd_sig(char *arg, char *error, size_t errorlen,
				 const char *signame, const char *varname)
{
	int ret;

	ret = nvl_set_str(current->qso->additional, varname, STR_DUP(arg));
	if (ret)
		snprintf(error, errorlen, "Failed to set rx %s reference: %s",
			 signame, xstrerror(ret));
	else
		snprintf(error, errorlen, "RX %s set to: %s", signame, arg);

	return ret;
}

static enum cmd_result cmd_pota(char *arg, char *error, size_t errorlen)
{
	return __cmd_sig(arg, error, errorlen, "POTA", "rx_pota");
}

static enum cmd_result cmd_wwff(char *arg, char *error, size_t errorlen)
{
	return __cmd_sig(arg, error, errorlen, "WWFF", "rx_wwff");
}

static void print_field(WINDOW *win, int row, int col, const char *str,
			size_t width, bool dirty)
{
	ssize_t len;

	wmove(win, row, col);

	if (!str)
		str = "";

	/*
	 * HACK: count number of printable characters
	 *
	 * We get UTF-8 strings passed in, but we need to know how wide it
	 * will print.  We assume that combining characters don't exist, and
	 * just count the number of bytes < 0x80 because that's how many
	 * codepoints there are.
	 */
	len = utf8_is_valid_str(str, strlen(str));
	ASSERT3S(len, >=, 0);

	if (len <= width) {
		size_t i;

		/* replace leading spaces with '_' */
		for (i = 0; (i < width) && (str[i] == ' '); i++)
			wprintw(win, "_");

		wattrset(win, dirty ? A_BOLD : A_NORMAL);
		wprintw(win, "%s", &str[i]);
		wattrset(win, A_NORMAL);

		/* padd with '_' */
		for (i = len; i < width; i++)
			wprintw(win, "_");
	} else {
		/* truncate & append ellipsis */
		wattrset(win, dirty ? A_BOLD : A_NORMAL);
		wprintw(win, "%.*s…", width - 1, str);
		wattrset(win, A_NORMAL);
	}
}

static int print_label(WINDOW *win, int row, int col, const char *str)
{
	bool emph;
	int count;
	size_t i;

	if (!str)
		return 0;

	wmove(win, row, col);

	for (i = 0, emph = false, count = 0; str[i]; i++) {
		if (str[i] == '_') {
			emph = true;
			continue;
		}

		waddch(win, str[i] | COLOR_PAIR(CMD_HINT_COLOR) |
		       (emph ? A_UNDERLINE : A_NORMAL));

		count++;
		emph = false;
	}

	return count;
}

static void print_tab_labels(void)
{
	char title[32];
	bool first;
	size_t i;

	/* TODO: We should just redraw the top line as needed */
	whrule(header_win, 0, ENTRY_WIN_WIDTH);

	wmove(header_win, 0, 2);

	wprintw(header_win, "[");
	for (i = 0, first = true; i < ARRAY_LEN(states); i++) {
		const bool this = (current == &states[i]);
		struct str *call;

		if (!states[i].in_use)
			continue;

		call = qso_get_station_call(&states[i].qso->rx);

		if (call)
			strcpy_safe(title, str_cstr(call), sizeof(title));
		else
			strcpy(title, "???");

		str_putref(call);

		if (!first)
			wprintw(header_win, " ");

		wattrset(header_win, this ? A_REVERSE : A_NORMAL);
		wprintw(header_win, "%zu:%s", i, title);
		wattrset(header_win, A_NORMAL);

		first = false;
	}
	wprintw(header_win, "]");
}

static void setup_windows(void)
{
	size_t i;
	int ret;

	header_win = newwin(HEADER_WIN_HEIGHT, HEADER_WIN_WIDTH, HEADER_WIN_ROW, 0);

	entry_win = newwin(ENTRY_WIN_HEIGHT, ENTRY_WIN_WIDTH, ENTRY_WIN_ROW, 0);

	for (i = 0; i < ARRAY_LEN(entry_fields); i++) {
		const struct entry_field *entry = &entry_fields[i];

		print_label(entry_win, entry->row, entry->col_label,
			    entry->label);
	}

	/* TODO: print '-' to separate the times */

	print_tab_labels();

	call_win = newwin(CALL_WIN_HEIGHT, CALL_WIN_WIDTH, CALL_WIN_ROW, 0);

	recent_win = newwin(RECENT_WIN_HEIGHT, RECENT_WIN_WIDTH, RECENT_WIN_ROW, 0);
	whrule(recent_win, 0, RECENT_WIN_WIDTH);
	whrule(recent_win, RECENT_WIN_HEIGHT - 1, RECENT_WIN_WIDTH);

	ret = xform_alloc(&cmdform, CMD_WIN_HEIGHT, CMD_WIN_WIDTH, CMD_WIN_ROW,
			  0, NULL, cmdfields, ARRAY_LEN(cmdfields));
	ASSERT0(ret);

	ret = status_init(STATUS_WIN_HEIGHT, STATUS_WIN_WIDTH, STATUS_WIN_ROW, 0,
			  0);
	ASSERT0(ret);
}

static void refresh_windows(void)
{
	wrefresh(header_win);
	wrefresh(entry_win);
	wrefresh(call_win);
	wrefresh(recent_win);
	xform_refresh(&cmdform);
	status_refresh();

	/* move cursor to the command line */
	xform_cursor_warp(&cmdform);
}

static void print_qso_line(WINDOW *win, struct qso *qso)
{
	char tmp[CALL_WIN_WIDTH];

	qso_line_fmt(qso, tmp, sizeof(tmp));

	wprintw(win, "%s", tmp);
}

static size_t __get_band(uint64_t band)
{
	size_t i;

	for (i = 0; i < ARRAY_LEN(summary_bands); i++)
		if (summary_bands[i] == band)
			return i;

	/* default to unknown/other column */
	ASSERT(!summary_bands[0]);

	return 0;
}

static size_t __get_mode(struct stats_summary *summary, struct str *mode)
{
	size_t min;
	size_t i;

	/* find matching mode name */
	for (i = 0; i < ARRAY_LEN(summary->mode); i++) {
		if (!summary->mode[i].total)
			continue;

		if ((summary->mode[i].mode_name && !mode) ||
		    (!summary->mode[i].mode_name && mode))
			continue;

		if (mode && str_cmp(summary->mode[i].mode_name, mode))
			continue;

		return i;
	}

	/* find an unused spot, and use it */
	for (i = 0; i < ARRAY_LEN(summary->mode); i++) {
		if (summary->mode[i].total)
			continue;

		summary->mode[i].mode_name = mode;

		return i;
	}

	/* find unknown/other and use it */
	for (i = 0; i < ARRAY_LEN(summary->mode); i++) {
		ASSERT(summary->mode[i].total);

		if (summary->mode[i].mode_name)
			continue;

		return i;
	}

	/* find least used mode, and convert it to unknown/other */
	for (i = 1, min = 0; i < ARRAY_LEN(summary->mode); i++) {
		ASSERT(summary->mode[i].total);
		ASSERT(summary->mode[i].mode_name);

		if (summary->mode[i].total < summary->mode[min].total)
			min = i;
	}

	summary->mode[min].mode_name = NULL;

	return min;
}

static void summarize_stats(struct stats_summary *summary)
{
	struct index_iter_ctx ctx;
	struct index_stat *stat;
	size_t mode, band;
	struct str *call;

	memset(summary, 0, sizeof(*summary));

	call = qso_get_station_call(&current->qso->rx);

	for (stat = index_iter_call_stats_start(call, &ctx);
	     stat;
	     stat = index_iter_call_stats_next(&ctx)) {
		band = __get_band(stat->band);
		mode = __get_mode(summary, stat->mode);

		summary->mode[mode].total += stat->count;
		summary->mode[mode].band[band] += stat->count;
	}

	str_putref(call);
}

static void refresh_fields(enum refresh refresh)
{
	size_t i;

	for (i = 0; i < ARRAY_LEN(entry_fields); i++) {
		const struct entry_field *entry = &entry_fields[i];
		struct val *val;

		if (!(entry->refresh & refresh))
			continue;


		if (entry->get.state) {
			val = entry->get.state();
		} else if (entry->get.qso) {
			val = entry->get.qso(current->qso);
		} else if (entry->get.side) {
			struct qso *qso = current->qso;
			struct qso_side *side;

			side = entry->tx_side ? &qso->tx : &qso->rx;

			val = entry->get.side(side);
		} else {
			panic("missing field refresh get function for %s",
			      entry->label);
		}

		/* no value => clear the field */
		if (!val) {
			print_field(entry_win, entry->row,
				    entry->col_value, NULL,
				    entry->len,
				    !!(current->dirty & entry->refresh));
			continue;
		}

		/* value => display it */
		switch (val->type) {
			case VT_STR:
				print_field(entry_win, entry->row,
					    entry->col_value,
					    val_cstr(val),
					    entry->len,
					    !!(current->dirty & entry->refresh));
				break;
			case VT_INT: {
				char buf[22];

				/* default to the empty string */
				buf[0] = '\0';

				switch (entry->refresh) {
					case REF_DE_POWER:
						pretty_print_power(buf,
								   sizeof(buf),
								   val);
						break;
					case REF_FREQ:
						pretty_print_freq(buf,
								  sizeof(buf),
								  val);
						break;
					case REF_ZONE_ITU:
					case REF_ZONE_CQZ:
						if (val)
							snprintf(buf, sizeof(buf),
								 "%02"PRIu64, val->i);
						break;
					case REF_ZONE_DXCC:
						if (val)
							snprintf(buf, sizeof(buf),
								 "%03"PRIu64, val->i);
						break;
					case REF_DE:
					case REF_DE_GRID:
					case REF_DATE:
					case REF_TIME_START:
					case REF_TIME_END:
					case REF_CALL:
					case REF_MODE:
					case REF_RSTS:
					case REF_RSTR:
					case REF_NAME:
					case REF_QTH:
					case REF_GRID:
					case REF_COUNTRY:
					case REF_GEO_LATLON:
					case REF_GEO_AZIDIST:
					case REF_STATE:
					case REF_COUNTY:
					case REF_ZONE_CONT:
					case REF_COMMENT:
					case REF_CALL_WIN:
					case REF_RECENT_WIN:
						panic("impossible");
				}

				print_field(entry_win, entry->row,
					    entry->col_value,
					    buf, entry->len,
					    !!(current->dirty & entry->refresh));
				break;
			}
			case VT_NULL:
			case VT_SYM:
			case VT_BOOL:
			case VT_CONS:
			case VT_CHAR:
			case VT_BLOB:
			case VT_ARRAY:
			case VT_NVL:
			default:
				panic("unhandled entry value type: %d", val->type);
		}

		val_putref(val);
	}

	if (refresh & REF_CALL_WIN) {
		/* index -> band in mm */
		struct stats_summary summary;
		struct index_iter_ctx ctx;
		struct str *call;
		struct qso *qso;
		size_t i, j;
		int row;

		/* clear it */
		for (row = 1; row < CALL_WIN_HEIGHT; row++)
		     mvwprintw(call_win, row, 0, "%*s", CALL_WIN_WIDTH - 1, "");

		summarize_stats(&summary);

		/* print header */
		mvwprintw(call_win, 1, 7, " #");
		for (i = 0; i < ARRAY_LEN(summary_bands); i++) {
			const int col = 10 + i * 4;

			if (!summary_bands[i])
				mvwprintw(call_win, 1, col, "U/O");
			else if (summary_bands[i] > 1000)
				mvwprintw(call_win, 1, col, "%3.0f",
					  summary_bands[i] / 1000.);
			else
				mvwprintw(call_win, 1, col, "%2.0fc",
					  summary_bands[i] / 10.);
		}

		/* print summary */
		row = 2;
		for (i = 0; i < ARRAY_LEN(summary.mode); i++) {
			const char *mode_name;

			if (!summary.mode[i].total)
				continue;

			/* print the mode name */
			if (summary.mode[i].mode_name)
				mode_name = str_cstr(summary.mode[i].mode_name);
			else
				mode_name = "<U/O>";

			wmove(call_win, row, 0);
			wprintw_cstr(call_win, mode_name, 6);

			/* print the mode total count */
			if (summary.mode[i].total < 100)
				wprintw(call_win, "%2"PRIu64,
					summary.mode[i].total);
			else
				wprintw(call_win, ">>");

			/* print per band indicator */
			for (j = 0; j < ARRAY_LEN(summary.mode[i].band); j++) {
				const int col = 10 + j * 4 + 1;

				if (!summary.mode[i].band[j])
					continue;

				mvwprintw(call_win, row, col, "X");
			}

			row++;
		}

		row++; /* skip a row */

		/* for each contact, print a line */
		call = qso_get_station_call(&current->qso->rx);
		for (qso = index_iter_call_history_start(call, &ctx, false);
		     qso;
		     qso = index_iter_call_history_next(&ctx), row++) {
			wmove(call_win, row, 0);
			print_qso_line(call_win, qso);

			if (row >= (CALL_WIN_HEIGHT - 1))
				break;
		}
		str_putref(call);
	}

	if (refresh & REF_RECENT_WIN)
		refresh_recent_win(recent_win, RECENT_WIN_HEIGHT - 1,
				   RECENT_WIN_WIDTH, true);
}

static enum cmd_result parse_command(char *buf, char *error, size_t errorlen)
{
	char *cmdname;
	size_t i;

	/* strip whitespace from buf - the command name */
	buf = trim_ws(buf);

	cmdname = strsep(&buf, " \t");

	/* strip whitespace from buf - the argument */
	buf = trim_ws(buf);

	for (i = 0; i < ARRAY_LEN(commands); i++) {
		const struct command *cmd = &commands[i];
		enum cmd_result res;

		if (strcasecmp(cmdname, cmd->cmd))
			continue;

		if (cmd->arg_required && !(buf && strlen(buf))) {
			snprintf(error, errorlen, "command '%s' requires an "
				 "argument", cmd->cmd);
			return -EINVAL;
		}
		if (!cmd->arg_required && (buf && strlen(buf))) {
			snprintf(error, errorlen, "command '%s' doesn't take "
				 "arguments", cmd->cmd);
			return -EINVAL;
		}

		res = cmd->fxn(buf, error, errorlen);
		if (res)
			return res;

		refresh_fields(cmd->refresh);
		return CR_OK;
	}

	snprintf(error, errorlen, "unknown command '%s'", cmdname);
	return -EINVAL;
}

static void usage(const char *prog)
{
	fprintf(stderr, "HLog version %s (libjeffpc %s)\n\n", version_string,
		jeffpc_version);

	fprintf(stderr, "Usage: %s <path to data dir>\n", prog);
	exit(1);
}

static int load_index(void)
{
	int ret;

	printf("Loading existing QSOs...\n");
	ret = index_init(AT_FDCWD);
	printf("%s.\n", ret ? "error" : "done");

	if (ret)
		fprintf(stderr, "Error: failed to load existing entries: %s\n",
			xstrerror(ret));

	return ret;
}

static void print_stats(void)
{
	struct index_iter_ctx ctx;
	struct qso *qso;
	char error[128];
	lua_State *L;
	bool failed;
	int ret;

	printf("Global call statistics:\n");

	L = xlua_init_script("startup-stats.lua", 0, error, sizeof(error));
	if (IS_ERR(L)) {
		printf("Error: failed to initialize lua: %s\n", error);
		return;
	}

	/* iterate over all the QSOs */
	for (qso = index_iter_history_start(&ctx, true), failed = false;
	     qso;
	     qso = index_iter_history_next(&ctx)) {
		ret = xlua_qso_call(L, "summarize_qso", qso_getref(qso),
				    true, error, sizeof(error));
		if (ret) {
			printf("Error: failed to execute lua: %s\n", error);
			failed = true;
		}
	}

	if (failed)
		printf("Warning: incomplete stats due to errors\n");

	/* print stats */
	ret = xlua_call(L, "print_summary", 0, 0, error, sizeof(error));
	if (ret)
		printf("Error: failed to execute lua (line %d): %s\n",
		       __LINE__, error);

	lua_close(L);
}

static void state_free(void)
{
	size_t i;

	for (i = 0; i < ARRAY_LEN(states); i++) {
		qso_putref(states[i].qso);

		str_putref(states[i].date);
		str_putref(states[i].time_start);
		str_putref(states[i].time_end);
		str_putref(states[i].latlon);
		str_putref(states[i].azidist);
	}
}

static int state_init(void)
{
	size_t i;

	/* reset the state */
	for (i = 0; i < ARRAY_LEN(states); i++) {
		states[i].qso = qso_alloc();
		if (!states[i].qso)
			goto err;

		states[i].date = &default_date;
		states[i].time_start = &default_time;
		states[i].time_end = &default_time;
		states[i].latlon = &default_latlon;
		states[i].azidist = &default_azidist;

		states[i].in_use = false;
	}

	/* we always start with the first one */
	current = &states[0];
	current->in_use = true;

	return 0;

err:
	state_free();

	return -ENOMEM;
}

static int load_and_apply_config(void)
{
	size_t i;
	int ret;

	/* allocate the template qso */
	template = qso_alloc();
	if (!template)
		return -ENOMEM;

	/* allocate & initialize blank states */
	ret = state_init();
	if (ret)
		goto err;

	/* read the config */
	ret = load_config(AT_FDCWD, template, "normal");
	if (ret)
		goto err_states;

	/* apply the template to the states */
	for (i = 0; i < ARRAY_LEN(states); i++)
		qso_copy(states[i].qso, template);

	return 0;

err_states:
	state_free();

err:
	qso_putref(template);

	return ret;
}

static void fill_generic_error(char *buf, size_t bufsize, int res)
{
#define COPYSTR(v) strcpy_safe(buf, (v), bufsize);
	if (res < 0) {
		COPYSTR(xstrerror(res));
	} else {
		switch ((enum cmd_result) res) {
			case CR_OK:
				panic("impossible");
			case CR_TOO_MANY_TABS:
				snprintf(buf, bufsize, "cannot open more than "
					 "%d tabs", MAX_STATE_COUNT);
				break;
			case CR_INVAL_TAB_ID:
				COPYSTR("invalid tab ID");
				break;
			case CR_UNKNOWN_TAB_ID:
				COPYSTR("tab doesn't exist");
				break;
		}
	}
#undef COPYSTR
}

static void ui_dispatch(int ch)
{
	struct xform *form = &cmdform;
	char error[CMD_WIN_WIDTH];
	enum cmd_result res;

	error[0] = '\0'; /* empty the error string */

	if (ui_dispatch_generic(form, ch))
		return;

	switch (ch) {
		/* delete & backspace */
		/* FIXME: what's the proper check? 127 works */
		case KEY_BACKSPACE:
		case '\b':
		case 127:
			xform_driver(form, XFORM_REQ_DEL_PREV);
			break;

		case KEY_DC:
			xform_driver(form, XFORM_REQ_DEL_THIS);
			break;

		/* navigate between forms */
		case KEY_ENTER:
		case '\n': {
			const void *tmp;
			size_t len;
			char *cmd;

			tmp = xform_field_get(&cmdform, 0, &len);
			cmd = alloca(len + 1);
			strcpy_safe(cmd, tmp, len + 1);

			res = parse_command(cmd, error, sizeof(error));
			if (res != CR_OK) {
				if (!error[0])
					fill_generic_error(error, sizeof(error),
							   res);

				status_display_error("Error: %s", error);
				beep();
			} else {
				status_display_info("%s", error);
			}

			xform_field_clear(form, 0);
			break;
		}

		case '\x1b': /* escape */
			xform_field_clear(form, 0);
			break;

		default:
			if (!iscntrl(ch)) {
				xform_driver(form, ch);
			} else {
				beep();
			}
			break;
	}
}
int main(int argc, char **argv)
{
	int ret;

	setlocale(LC_ALL, "");
	ASSERT0(putenv("TZ=UTC"));

	if (argc != 2)
		usage(argv[0]);

	data_dir = argv[1];

	if (chdir(data_dir)) {
		fprintf(stderr, "Error: failed to chdir to data dir: %s\n",
			strerror(errno));
		goto err;
	}

	ret = load_databases();
	if (ret)
		goto err;

	ret = load_index();
	if (ret)
		goto err;

	print_stats();

	ret = load_and_apply_config();
	if (ret)
		goto err;

	initscr();
	start_color();

	init_pair(CMD_HINT_COLOR, COLOR_YELLOW, COLOR_BLACK);

	setup_windows();
	refresh_fields(~0u);
	refresh_windows();

	for (;;) {
		int ch;

		ch = xform_getch(&cmdform);
		if (ch != ERR)
			ui_dispatch(ch);

		print_tab_labels();

		refresh_windows();
	}

	xform_free(&cmdform);
	delwin(header_win);
	delwin(entry_win);
	delwin(call_win);
	delwin(recent_win);
	status_deinit();

	endwin();

	state_free();

	return 0;

err:
	return 1;
}
