/*
 * Copyright (c) 2021-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "xlua_impl.h"
#include <hlog/country.h>
#include <hlog/zones.h>

/* return a table with country info or nil [-2,+1,-] */
static int xlua_country_lookup(lua_State *L)
{
	struct country_result res;
	const char *state;
	const char *call;

	xlua_checkstack(L, 2, 2, 2);

	call = luaL_checkstring(L, -2);
	if (lua_isnil(L, -1))
		state = NULL;
	else
		state = luaL_checkstring(L, -1);

	/*
	 * First, get info from country data
	 */
	if (!country_lookup(call, &res)) {
		lua_pushnil(L);
		return 1;
	}

	lua_newtable(L);

#define _S(n, f, v) \
	do { \
		f(L, v); \
		lua_setfield(L, -2, n); \
	} while (0)
#define S(n, x) \
	do { \
		if (x) \
			_S(n, lua_pushstring, str_cstr(x)); \
	} while (0)
#define N(n, x) _S(n, lua_pushnumber, x)
#define I(n, x, allowzero) \
	do { \
		if (allowzero || (x)) \
			_S(n, lua_pushinteger, x); \
	} while (0)

	S("prefix", res.prefix);
	S("name", res.name);
	S("country", res.fullname);
	S("continent", res.continent);
	N("lat", res.lat);
	N("lon", res.lon);
	I("tz", res.tz, true);
	I("dxcc", res.code, true);

	/* some DXCC entities need a state to determine CQ/ITU zones */
	switch (res.code) {
		case 1: /* Canada */
		case 6: /* Alaska */
		case 110: /* Hawaii */
		case 291: /* United States */
			break;
		default:
			I("cqz", res.cqz, false);
			I("itu", res.itu, false);
			break;
	}

	str_putref(res.prefix);
	str_putref(res.name);
	str_putref(res.fullname);
	str_putref(res.continent);

	/*
	 * Second, override CQ and ITU zones based on state
	 */
	if (state) {
		int ret;

		ret = cqz_by_state(res.code, state);
		if (ret >= 0)
			I("cqz", ret, false);

		ret = ituz_by_state(res.code, state);
		if (ret >= 0)
			I("itu", ret, false);
	}

	return 1;
}

static const luaL_Reg country_funcs[] = {
	{ "lookup", xlua_country_lookup },
	{ NULL, NULL },
};

/* register the country library [-0,+0,e] */
void xlua_pushlib_country(lua_State *L)
{
	luaL_newlib(L, country_funcs);
}
