/*
 * Copyright (c) 2020-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef __XLUA_H
#define __XLUA_H

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#include <jeffpc/error.h>
#include <jeffpc/cstr.h>

#include <hlog/qso.h>

/* make sure the stack is empty */
#define ASSERT_LUA_STACK(L, orig) \
	do { \
		int _n = lua_gettop(L); \
		int _o = (orig); \
		if (_n != _o) { \
			xlua_dump_stack(L); \
			panic("lua stack for %p has %d items, %d expected", \
			      (L), _n, _o); \
		} \
	} while (0)

/* copy error from stack to string */
#define COPY_ERROR(L, idx, err, errlen) \
	strcpy_safe(err, lua_tostring((L), (idx)), (errlen))

extern int lua2errno(int e);
extern lua_State *xlua_init_at(int dirfd, const char *script_fname, int nresults,
			       char *error, size_t errlen);
extern lua_State *xlua_init_script(const char *script_name, int nresults,
				   char *error, size_t errlen);

/* return raw script data */
extern int xlua_script_find(const char *name, const uint8_t **data_r,
			    size_t *size_r);

/*
 * Call a function.
 *
 * If the function name is NULL, the item just before the arguments is
 * called.
 */
extern int xlua_call(lua_State *L, const char *funcname, int nargs, int nresults,
		     char *error, size_t errlen);
/*
 * NULL function name implies that the function is at the top of the stack;
 * consumes the passed in qso ref
 */
extern int xlua_qso_call(lua_State *L, const char *funcname, struct qso *qso,
			 bool ro, char *error, size_t errlen);
extern void xlua_dump_stack(lua_State *L);

/*
 * Push wrapped C types
 */
/* the following consume the passed in reference */
extern int xlua_pushnvl(lua_State *L, struct nvlist *nvl, bool ro,
			char *error, size_t errlen);
extern int xlua_pushqso(lua_State *L, struct qso *qso, bool ro,
			char *error, size_t errlen);
/* the following aren't reference counted */
extern int xlua_pushtime(lua_State *L, struct time *time, bool ro,
			 char *error, size_t errlen);

static inline lua_State *xlua_init(const char *script_fname, int nresults,
				   char *error, size_t errlen)
{
	return xlua_init_at(AT_FDCWD, script_fname, nresults, error, errlen);
}

static inline bool xlua_isstringornil(lua_State *L, int idx)
{
	return lua_isnil(L, idx) || lua_isstring(L, idx);
}

#endif
