--
-- Copyright (c) 2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--

local allowed_bands = {
	[80000] = true,
	[40000] = true,
	[20000] = true,
	[15000] = true,
	[10000] = true,
}

-- DXCC entities that use states (+counties) instead of "DX"
local uscan_dxcc = {
	[1] = true, -- Canada
	[6] = true, -- Alaska
	[110] = true, -- Hawaii
	[291] = true, -- CONUS
}

local raw_county_data = {
	{ 'CT', 'FAI', 'Fairfield' },		{ 'CT', 'HAR', 'Hartford' },
 	{ 'CT', 'LIT', 'Litchfield' },		{ 'CT', 'MID', 'Middlesex' },
 	{ 'CT', 'NHV', 'New-Haven' },		{ 'CT', 'NLN', 'New-London' },
 	{ 'CT', 'TOL', 'Tolland' },		{ 'CT', 'WIN', 'Windham' },

	{ 'MA', 'BAR', 'Barnstable' },		{ 'MA', 'BER', 'Berkshire' },
 	{ 'MA', 'BRI', 'Bristol' },		{ 'MA', 'DUK', 'Dukes' },
 	{ 'MA', 'ESS', 'Essex' },		{ 'MA', 'FRA', 'Franklin' },
 	{ 'MA', 'HMD', 'Hampden' },		{ 'MA', 'HMP', 'Hampshire' },
 	{ 'MA', 'MID', 'Middlesex' },		{ 'MA', 'NAN', 'Nantucket' },
 	{ 'MA', 'NOR', 'Norfolk' },		{ 'MA', 'PLY', 'Plymouth' },
 	{ 'MA', 'SUF', 'Suffolk' },		{ 'MA', 'WOR', 'Worcester' },

	{ 'ME', 'AND', 'Androscoggin' },	{ 'ME', 'ARO', 'Aroostook' },
 	{ 'ME', 'CUM', 'Cumberland' },		{ 'ME', 'FRA', 'Franklin' },
 	{ 'ME', 'HAN', 'Hancock' },		{ 'ME', 'KEN', 'Kennebec' },
 	{ 'ME', 'KNO', 'Knox' },		{ 'ME', 'LIN', 'Lincoln' },
 	{ 'ME', 'OXF', 'Oxford' },		{ 'ME', 'PEN', 'Penobscot' },
 	{ 'ME', 'PIS', 'Piscataquis' },		{ 'ME', 'SAG', 'Sagadahoc' },
 	{ 'ME', 'SOM', 'Somerset' },		{ 'ME', 'WAL', 'Waldo' },
 	{ 'ME', 'WAS', 'Washington' },		{ 'ME', 'YOR', 'York' },

	{ 'NH', 'BEL', 'Belknap' },		{ 'NH', 'CAR', 'Carroll' },
 	{ 'NH', 'CHE', 'Cheshire' },		{ 'NH', 'COO', 'Coos' },
 	{ 'NH', 'GRA', 'Grafton' },		{ 'NH', 'HIL', 'Hillsborough' },
 	{ 'NH', 'MER', 'Merrimack' },		{ 'NH', 'ROC', 'Rockingham' },
 	{ 'NH', 'STR', 'Strafford' },		{ 'NH', 'SUL', 'Sullivan' },

	{ 'RI', 'BRI', 'Bristol' },		{ 'RI', 'KEN', 'Kent' },
 	{ 'RI', 'NEW', 'Newport' },		{ 'RI', 'PRO', 'Providence' },
 	{ 'RI', 'WAS', 'Washington' },

	{ 'VT', 'ADD', 'Addison' },		{ 'VT', 'BEN', 'Bennington' },
 	{ 'VT', 'CAL', 'Caledonia' },		{ 'VT', 'CHI', 'Chittenden' },
 	{ 'VT', 'ESS', 'Essex' },		{ 'VT', 'FRA', 'Franklin' },
 	{ 'VT', 'GRA', 'Grand Isle' },		{ 'VT', 'LAM', 'Lamoille' },
 	{ 'VT', 'ORA', 'Orange' },		{ 'VT', 'ORL', 'Orleans' },
 	{ 'VT', 'RUT', 'Rutland' },		{ 'VT', 'WAS', 'Washington' },
 	{ 'VT', 'WNH', 'Windham' },		{ 'VT', 'WND', 'Windsor' },
}

-- generate views of the county data
local map_county_name2abbrev = {}
local map_county_abbrev2name = {}
local new_england_states = {}

-- initialize county_{name2abbrev,abbrev2name}
for _, info in pairs(raw_county_data) do
	local a = info[2] .. info[1]
	local n = info[3] .. info[1]
	local s = info[1]

	assert(map_county_name2abbrev[n] == nil)
	map_county_name2abbrev[n] = a

	assert(map_county_abbrev2name[a] == nil)
	map_county_abbrev2name[a] = info

	new_england_states[s] = true
end

local function county_name2abbrev(dxcc, state, county)
	-- outside of US or Canada: DX
	if uscan_dxcc[dxcc] == nil then
		return "DX"
	end

	-- outside of New England: state/province
	if new_england_states[state] == nil then
		return state
	end

	-- inside New England: <county><state> abbreviation
	assert(county ~= nil)

	return map_county_name2abbrev[county .. state]
end

local function county_abbrev2name(abbrev)
	local info = map_county_abbrev2name[abbrev]

	-- inside New England: return state & county
	if info ~= nil then
		return info[1], info[3]
	end

	-- outside of New England: state
	if abbrev ~= nil and abbrev ~= "DX" then
		return abbrev, nil
	end

	-- outside of US
	return nil, nil
end

local function qso_done(qso)
	-- nothing to do for now
end

-- copy country data into QSO
local function sync_zones(qso)
	local call = qso.rx.station_call
	local state = qso.rx.state
	local cty = nil

	if call ~= nil then
		cty = hlog.country.lookup(call, state)
	end

	if cty == nil then
		cty = {}
	end

	qso.rx.dxcc = cty.dxcc
	qso.rx.itu = cty.itu
	qso.rx.cqz = cty.cqz
	qso.rx.country = cty.country
	qso.rx.continent = cty.continent
end

local function field_changed(qso, fields, field)
	local value = fields[field]

	if field == "rx.station_call" then
		qso.rx.station_call = value
		sync_zones(qso)
	elseif field == "tx.rst" then
		qso.tx.rst = value
	elseif field == "rx.rst" then
		qso.rx.rst = value
	elseif field == "rx.exch" then
		-- exchange should always be uppercased
		value = value:upper()

		-- store raw exchange
		qso.additional['contest-rx'] = value

		-- set state & county based on the exchange
		local state, county = county_abbrev2name(value)
		qso.rx.state = state
		qso.rx.county = county

		-- update zones now that we may have a state set
		sync_zones(qso)
	elseif field == "comment" then
		qso.comment = value
	else
		error(string.format("Unknown field '%s' encountered", field))
	end
end

return {
	labels = {
		{ 0,  0, "St. Call" },
		{ 0, 11, "RSTs" },
		{ 0, 16, "RSTr" },
		{ 0, 21, "Exchange" },
		{ 0, 30, "Comment" },
	},

	fields = {
		{ 1,  0, 10, "rx.station_call", },
		{ 1, 11,  4, "rx.rst", },
		{ 1, 16,  4, "rx.rst", },
		{ 1, 21,  8, "rx.exch", },
		{ 1, 30, 24, "comment", },
	},

	events = {
		-- called when QSO is being initialized
		--
		-- We rely on the dxcc, state, county being set by the
		-- template.
		qso_init = function(qso)
			local dxcc = qso.tx.dxcc
			local state = qso.tx.state
			local county = qso.tx.county
			local cs = county_name2abbrev(dxcc, state, county)

			qso.tx.rst = "59" -- we always send 59
			qso.additional['contest-id'] = 'NEQP'
			qso.additional['contest-year'] = 2022

			if cs == nil then
				contest.lcd("?")
			else
				qso.additional['contest-tx'] = cs
				contest.lcd(cs)
			end
		end,

		-- called when QSO is complete & saved, just prior to a reset
		qso_done = qso_done,

		-- called whenever a contest field changes
		field_changed = field_changed,
	},
}
