--
-- Copyright (c) 2021-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--

local HIST_HEIGHT = 10

local us_states = {
	"AK", "AL", "AR", "AZ",
	"CA", "CO", "CT",
	"DC", "DE",
	"FL",
	"GA",
	"HI",
	"IA", "ID", "IL", "IN",
	"KS", "KY",
	"LA",
	"MA", "MD", "ME", "MI", "MN", "MO", "MS", "MT",
	"NC", "ND", "NE", "NH", "NJ", "NM", "NV", "NY",
	"OH", "OK", "OR",
	"PA",
	"RI",
	"SC", "SD",
	"TN", "TX",
	"UT",
	"VA", "VT",
	"WA", "WI", "WV", "WY",
}

local ca_provinces = {
	-- actual provinces
	"AB",
	"BC",
	"MB",
	"NB", "NL", "NS",
	"ON",
	"PE",
	"QC",
	"SK",

	-- territories
	"NT", "NU",
	"YT",
}

--
-- Utils
--

local function padnum(v, width)
	local tmp = tostring(v)

	while tmp:len() < width do
		tmp = " " .. tmp
	end

	return tmp
end

local function padnum_max(c, m)
	c = tostring(c)
	m = tostring(m)

	if c:len() < m:len() then
		return string.rep(" ", m:len() - c:len()) .. c
	end

	return c
end

--
-- Contest
--

-- The result of a lookup of the park we're in.
local my_park_info

-- This contains <hash key, dummy value> so we can quickly check for
-- duplicates.
local dup_table = {}

local function hash(qso, txref, rxref)
	local rxstate = qso.rx.state
	local band = (qso.tx.band ~= nil) and (qso.tx.band / 1000) or "NONE"
	local mode = qso.tx.mode

	return string.format("%04u%02u%02u_%s-%s-%s_%s-%s-%s_%s-%s",
			     -- when
			     qso.start_time.year,
			     qso.start_time.month,
			     qso.start_time.day,
			     -- me
			     qso.tx.station_call,
			     qso.tx.state,
			     txref,
			     -- other op
			     qso.rx.station_call,
			     (rxstate ~= nil and rxref ~= nil) and rxstate or "NONE",
			     (rxref ~= nil) and rxref or "NONE",
			     -- contact
			     band,
			     mode)
end

local function get_hashes(qso)
	local txrefs = hlog.utils.split_on_commas(qso.additional['tx_pota'])
	local rxrefs = hlog.utils.split_on_commas(qso.additional['rx_pota'])

	local out = {}
	for _, txref in ipairs(txrefs) do
		if rxrefs == nil then
			-- regular hunter
			table.insert(out, hash(qso, txref, nil))
		else
			-- Park to Park contact, handle the other activator
			-- being in multiple entities concurrently
			for _, rxref in ipairs(rxrefs) do
				table.insert(out, hash(qso, txref, rxref))
			end
		end
	end

	return out
end

local function qso_done(qso)
	for _, h in ipairs(get_hashes(qso)) do
		dup_table[h] = true
	end
end

local function isdup(qso)
	local hashes = get_hashes(qso)
	for _, h in ipairs(hashes) do
		if dup_table[h] == nil then
			-- at least one hash doesn't exist in the list of
			-- known QSOs
			return false
		end
	end

	return true
end

-- returns true if QSO counts for early/late shift award
local function isearlylate(early, qso)
	local hour = qso.start_time.hour
	local stime, etime

	if early then
		stime = my_park_info['early_shift_start_time']
		etime = my_park_info['early_shift_end_time']
	else
		stime = my_park_info['late_shift_start_time']
		etime = my_park_info['late_shift_end_time']
	end

	if stime < etime then
		-- no end-of-day wrap
		return (stime <= hour) and (hour < etime)
	else
		-- end of day wrap
		return (stime <= hour) or (hour < etime)
	end
end

local function copy_if_not_nil(dst, dkey, src)
	if src ~= nil then
		dst[dkey] = src
	end
end

-- copy country data into QSO
--
-- If the station callsign is nil (i.e., empty), we clear the various
-- QSO location fields.  In theory, we could clear the fields whenever we
-- get a lookup failure, but that seems a bit dangerous.
local function sync_zones(qso)
	local call = qso.rx.station_call
	local state = qso.rx.state

	if call ~= nil then
		local cty = hlog.country.lookup(call, state)
		if cty == nil then
			return
		end

		copy_if_not_nil(qso.rx, 'dxcc', cty.dxcc)
		copy_if_not_nil(qso.rx, 'itu', cty.itu)
		copy_if_not_nil(qso.rx, 'cqz', cty.cqz)
		copy_if_not_nil(qso.rx, 'country', cty.country)
		copy_if_not_nil(qso.rx, 'continent', cty.continent)
	else
		qso.rx.dxcc = nil
		qso.rx.itu = nil
		qso.rx.cqz = nil
		qso.rx.country = nil
		qso.rx.continent = nil
	end
end

local function field_changed(qso, fields, field)
	local value = fields[field]

	if field == "rx.station_call" then
		qso.rx.station_call = value
		sync_zones(qso)
	elseif field == "tx.rst" then
		qso.tx.rst = value
	elseif field == "rx.rst" then
		qso.rx.rst = value
	elseif field == "rx.state" then
		qso.rx.state = value:upper()
		sync_zones(qso)
	elseif field == "name" then
		qso.rx.name = value
	elseif field == "ref" then
		qso.additional.rx_pota = value:upper()

		local res = hlog.sig.lookup_pota(value:upper())
		if res ~= nil then
			qso.rx.qth = res['name']
		end

		-- TODO: fill in more information (state, cqz, ituz, ...)
	else
		error(string.format("Unknown field '%s' encountered", field))
	end

	return nil, {
		["DUP"] = isdup(qso) and "red" or "black",
		["AM"] = isearlylate(true, qso) and "green" or "black",
		["PM"] = isearlylate(false, qso) and "green" or "black",
	}
end

local function startup_fill_template(template)
	local function yesno(msg, default)
		local other = {
			['y'] = 'n',
			['n'] = 'y',
		}
		other = other[default:lower()]

		print(msg .. " [" .. default:upper() .. other .. "]")

		local ack = io.stdin:read()
		if ack == other:lower() or ack == other:upper() then
			return true
		end

		return false
	end

	local function get_state(dxcc, locs)
		local function strip_pfx(loc, pfx)
			return loc:sub(pfx:len() + 1)
		end

		local pfxs = {
			[1] = "CA-", -- Canada
			[6] = "US-", -- Alaska
			[110] = "US-", -- Hawaii
			[291] = "US-", -- continental US
		}

		if pfxs[dxcc] == nil then
			return nil
		end

		if locs:find(",") == nil then
			return strip_pfx(locs, pfxs[dxcc])
		end

		while true do
			print("Which state (e.g., US-MA)?")
			local tmp = io.stdin:read()

			for _, loc in ipairs(hlog.utils.split_on_commas(locs)) do
				assert(loc:find(pfxs[dxcc]) == 1)

				local state = strip_pfx(loc, pfxs[dxcc])

				if tmp == loc or tmp == state then
					return state
				end
			end

			if yesno("Entered state not listed, use anyway?", 'n') then
				return tmp
			end
		end
	end

	local park
	local state

	repeat
		print()
		print("Enter park reference (e.g., K-1234):")

		local ref = io.stdin:read()
		if ref == "" then
			if yesno("Empty park reference, use it anyway?", 'n') then
				return
			end
		else
			park = hlog.sig.lookup_pota(ref)

			if park == nil then
				if yesno("Unknown park reference, use it anyway?", 'n') then
					-- We know nothing but the reference
					template.additional['tx_pota'] = ref
					return
				end
			else
				print("Found:")
				for k,v in pairs(park) do
					print("", k, v)
				end
				print()
				if not park['active'] then
					print("Warning: this reference is marked as inactive!")
					print()
				end

				state = get_state(park['dxcc'], park['locs'])

				if yesno("Use it?", 'y') then
					park = nil
				end
			end
		end
	until park ~= nil

	template.tx.qth = park['name']
	template.tx.dxcc = park['dxcc']
	template.tx.state = state
	template.additional['tx_pota'] = park['ref']

	-- TODO: fill in country, continent, ITU, CQ

	-- pre-calculate the early and late shift start/end times
	if park['lat'] ~= nil and park['lon'] ~= nil and #hlog.utils.split_on_commas(park['locs']) == 1 then
		local function calc(start, duration)
			-- from award description on parksontheair.com
			local s = hlog.utils.round(start - (park['lon'] / 15))

			-- make sure the start time is [0, 24)
			while s < 0 do
				s = s + 24
			end
			while s >= 24 do
				s = s - 24
			end

			local e = s + duration

			-- make sure the end time is [0, 24)
			while e < 0 do
				e = e + 24
			end
			while e >= 24 do
				e = e - 24
			end

			return s, e
		end

		-- early shift: starts at 02:00 UTC and lasts 6 hours
		park['early_shift_start_time'], park['early_shift_end_time'] = calc(2, 6)

		-- late shift: starts at 18:00 UTC and lasts 8 hours
		park['late_shift_start_time'], park['late_shift_end_time'] = calc(18, 8)
	end

	-- stash park info for later reference
	my_park_info = park
end

--
-- Export
--

-- Generate a unique key for a contact.
local function gen_key(qso)
	local call = qso.rx.station_call
	local band = qso.tx.band / 1000
	local mode = qso.tx.mode
	local rxref = qso.additional['sig_info']

	return string.format("%s-%s-%s-%s", call, band, mode,
			     (rxref ~= nil) and rxref or "NONE")
end

-- Drop the various "mobile" and "portable" suffixes.  Not only does POTA
-- not require them, it gets a little confused by them.  Specifically, it'll
-- break the activity/activation into one per station call.  This can be
-- worked around in the POTA UI by adding additional calls to your account,
-- but that's annoying.  So, we pretend the call was without one of these
-- common and trivial suffixes.
local function simplify_station_call(call)
	for _, suffix in ipairs({ "P", "M", "MM", "AM" }) do
		if call:match("/"..suffix.."$") ~= nil then
			return call:sub(1, -2 - suffix:len())
		end
	end

	return call
end

local function do_add_entry(prog, signame, station, orig_qso, txref, rxref)
	local basename = string.format("%s_%s_%s_%04u%02u%02u",
				       orig_qso.tx.operator_call,
				       txref:gsub('/', '-'),
				       orig_qso.tx.state,
				       orig_qso.start_time.year,
				       orig_qso.start_time.month,
				       orig_qso.start_time.day)

	-- get a writable copy
	qso = orig_qso:clone()
	hlog.nvl.clear(qso.additional)
	if signame == "SOTA" then
		qso.additional['my_sota_ref'] = txref
		qso.additional['sota_ref'] = rxref
	else
		if txref ~= nil then
			qso.additional['my_sig'] = signame
			qso.additional['my_sig_info'] = txref
		end
		if rxref ~= nil then
			qso.additional['sig'] = signame
			qso.additional['sig_info'] = rxref
		end
	end

	-- drop /P, /M, etc. from station call
	qso.tx.station_call = station

	-- clear undesired qso & side fields
	qso.comment = nil

	if qso.rx.freq == qso.tx.freq then
		qso.rx.freq = nil
	end
	if qso.rx.mode == qso.tx.mode then
		qso.rx.mode = nil
	end

	local cur = prog.logs[basename]
	if cur == nil then
		cur = {
			chrono = {}, -- array of QSOs, chronologically ordered
			dedup = {}, -- <key,count> dedup counter
			states = {}, -- <state,count> rx states stats
			dxcc = {}, -- <dxcc,count> rx dxcc stats
			p2p = {}, -- <call,rx ref> park-to-park stats
			bandmode = {}, -- <band,<mode,count>>
		}
		prog.logs[basename] = cur
	end

	-- We don't have to include date & tx ref, since those are already
	-- taken care of.
	local key = gen_key(qso)

	-- add QSO to chronological array
	table.insert(cur.chrono, qso)
	prog.mangled_qso_to_orig[qso] = orig_qso

	-- account for QSO in dedup map
	cur.dedup[key] = (cur.dedup[key] ~= nil) and (cur.dedup[key] + 1) or 1

	-- account for QSO dxcc
	if qso.rx.dxcc ~= nil then
		local dxcc = string.format("%s (%s)", qso.rx.country, qso.rx.dxcc)

		cur.dxcc[dxcc] = (cur.dxcc[dxcc] ~= nil) and (cur.dxcc[dxcc] + 1) or 1
	end

	-- account for QSO state
	if qso.rx.state ~= nil then
		local st = qso.rx.state

		cur.states[st] = (cur.states[st] ~= nil) and (cur.states[st] + 1) or 1
	end

	-- account for park-to-park stats
	if rxref ~= nil then
		if cur.p2p[qso.rx.station_call] == nil then
			cur.p2p[qso.rx.station_call] = {}
		end

		table.insert(cur.p2p[qso.rx.station_call], rxref)
	end

	-- account for band & mode
	local mode = qso.tx.mode
	local band = qso.tx.band_name

	if cur.bandmode[band] == nil then
		cur.bandmode[band] = {}
	end
	if cur.bandmode[band][mode] == nil then
		cur.bandmode[band][mode] = 0
	end
	cur.bandmode[band][mode] = cur.bandmode[band][mode] + 1

	--
	-- Finally, clean up the QSO struct, *after* we did our stats
	-- accounting.
	--
	for _, key in ipairs({ "name", "power", "qth", "country", "county",
			       "continent", "dxcc", "itu", "cqz", "seq" }) do
		qso.tx[key] = nil
		qso.rx[key] = nil
	end
end

local function add_entry(qso, prog, signame, tx, rx, submit, allow_multiple)
	-- needs to be a contact that we haven't exported before or are
	-- labeled as bad
	if (qso.additional[tx] == nil) or
	   (qso.additional[submit] ~= nil) or
	   (qso.additional['bad'] ~= nil) then
		return
	end

	if qso.tx.band == nil then
		print(string.format("Error: %s QSO lacks band info", qso.uuid))
		return
	end

	assert(qso.tx.station_call ~= nil)

	local station = simplify_station_call(qso.tx.station_call)

	-- TODO: check for dups

	local txrefs = hlog.utils.split_on_commas(qso.additional[tx])
	local rxrefs = hlog.utils.split_on_commas(qso.additional[rx])

	assert(txrefs ~= nil)

	if not allow_multiple and hlog.utils.count_items(txrefs) > 1 then
		print(string.format("Warning: %s QSO with multiple tx %s references (%s)",
				    qso.uuid, signame, qso.additional[tx]))
	end
	if not allow_multiple and hlog.utils.count_items(rxrefs) > 1 then
		print(string.format("Warning: %s QSO multiple rx %s references (%s)",
				    qso.uuid, signame, qso.additional[rx]))
	end

	for _, txref in ipairs(txrefs) do
		if rxrefs == nil then
			-- regular hunter
			do_add_entry(prog, signame, station, qso, txref, nil)
		else
			-- P2P contact, handle the other activator
			-- being in multiple entities concurrently
			for _, rxref in ipairs(rxrefs) do
				do_add_entry(prog, signame, station, qso, txref, rxref)
			end
		end
	end
end

local function write_dxcc_stats(f, dxccs)
	f:write("DXCC stats:\n")
	f:write("  Count Country (DXCC)\n")
	local by_count = {}
	for dxcc, count in pairs(dxccs) do
		local tmp = by_count[count]
		if tmp == nil then
			by_count[count] = {}
		end
		table.insert(by_count[count], dxcc)
	end

	for _, count in ipairs(hlog.utils.sorted_keys(by_count)) do
		local dxccs = by_count[count]
		table.sort(dxccs)

		for _, dxcc in ipairs(dxccs) do
			f:write(string.format("  %s %s\n", padnum(count, 5), dxcc))
		end
	end
	f:write("\n")
end

local function write_state_stats(f, states)
	local function count_states(allstates)
		local count = 0
		local maxstate = 0
		for _, state in ipairs(allstates) do
			if states[state] then
				count = count + 1
				maxstate = math.max(maxstate, states[state])
			end
		end

		return count, maxstate
	end

	local function collate_states(allstates, maxstate)
		local got = {}
		local mis = {}
		local hist = {}
		for _, state in ipairs(allstates) do
			if states[state] ~= nil then
				local hashes = math.ceil(HIST_HEIGHT * states[state] / maxstate)
				local blanks = HIST_HEIGHT - hashes

				table.insert(got, string.format("%s(%u)", state, states[state]))
				table.insert(hist,
					     string.rep(" ", blanks) ..
					     string.rep("#", hashes) ..
					     "-" ..
					     state ..
					     padnum_max(states[state], maxstate))
			else
				table.insert(mis, string.format("%s", state))
				table.insert(hist,
					     string.rep(" ", HIST_HEIGHT) ..
					     "-" ..
					     state ..
					     string.rep(" ", tostring(maxstate):len() - 1).."-")
			end
		end

		return got, mis, hist
	end

	local function list_states(pfx, list)
		local out = "  " .. pfx
		local neednl = true

		for idx, state in ipairs(list) do
			out = out .. " " .. state
			neednl = true

			if out:len() > 60 then
				f:write(out .. "\n")
				out = "       "
				neednl = false
			end
		end

		if neednl then
			f:write(out .. "\n")
		end
	end

	local function print_summary(label, count, allstates, got, mis)
		f:write(string.format("%s stats: %2u/%2u\n", label, count,
				      hlog.utils.count_items(allstates)))

		list_states("Got: ", got)
		list_states("Need:", mis)
	end

	-- find number of states and most frequent state
	local us_count, maxstate1 = count_states(us_states)
	local ca_count, maxstate2 = count_states(ca_provinces)
	local maxstate = math.max(maxstate1, maxstate2)

	if count == 0 then
		return -- don't print state stats if we have none
	end

	-- output state counts
	us_got, us_mis, us_hist = collate_states(us_states, maxstate)
	ca_got, ca_mis, ca_hist = collate_states(ca_provinces, maxstate)

	print_summary("State", us_count, us_states, us_got, us_mis)
	f:write("\n")
	print_summary("Province/territory", ca_count, ca_provinces, ca_got, ca_mis)

	-- histogram
	f:write(string.format("\n  Each '#' = %.1f contacts:\n", maxstate / HIST_HEIGHT))
	for row = 1, HIST_HEIGHT + 1 + 2 + tostring(maxstate):len() do
		local out = "  "
		for _, h in ipairs(us_hist) do
			out = out .. h:sub(row, row)
		end
		out = out .. " "
		for _, h in ipairs(ca_hist) do
			out = out .. h:sub(row, row)
		end
		f:write(out .. "\n")
	end
end

local function write_bandmode_stats(f, cur)
	-- get a sorted list of all subkeys
	local allb = {}
	local widths = {}
	for a, _ in pairs(cur) do
		for b, c in pairs(cur[a]) do
			allb[b] = true
			widths[b] = math.max(b:len(), tostring(c):len())
		end
	end
	allb = hlog.utils.sorted_keys(allb)

	-- print header
	local out = string.format("  %4s ", "")
	for _, b in ipairs(allb) do
		out = out .. string.format("%s ", b)
	end
	f:write(out .. "| Total\n")

	-- print data rows
	for _, a in ipairs(hlog.utils.sorted_keys(cur)) do
		local total = 0
		local out = string.format("  %-4s ", a)
		for _, b in ipairs(allb) do
			if cur[a][b] ~= nil then
				total = total + cur[a][b]
				out = out .. string.format("%s ", padnum(cur[a][b], widths[b]))
			else
				out = out .. string.format("%s ", padnum("", widths[b]))
			end
		end

		f:write(string.format("%s| %s\n", out, padnum(total, 5)))
	end
end

local function open_uuid_file(f, basename, suffix)
	if f ~= nil then
		return f
	end

	return assert(io.open(basename .. suffix, "w"))
end

local function close_uuid_file(f)
	if f ~= nil then
		f:close()
	end
end

local function gen_logs(prog, progname, outdir)
	local logs = prog.logs

	for basename,cur in pairs(logs) do
		print(basename)

		-- generate files with all the uuids submitted
		local logpath = outdir .. "/" .. basename
		local uuids = nil
		local f = assert(io.open(logpath .. ".adif", "w"))

		-- Statistics
		f:write(string.format("%s ADIF export - %s\n",
				      progname, basename))

		-- summary
		f:write(string.format("%u contacts, %u unique, %u dups\n",
				      hlog.utils.count_items(cur.chrono),
				      hlog.utils.count_items(cur.dedup),
				      hlog.utils.count_items(cur.chrono) -
				      hlog.utils.count_items(cur.dedup)))

		-- duplicates
		for key, count in pairs(cur.dedup) do
			assert(count ~= 0)

			if count > 1 then
				f:write(string.format("%u dups for %s\n", count - 1, key))
			end
		end
		f:write("\n")

		-- bands & modes
		f:write("Bands & modes:\n")
		write_bandmode_stats(f, cur.bandmode)
		f:write("\n")

		-- park-to-park stats
		if hlog.utils.count_items(cur.p2p) > 0 then
			f:write("Park-to-park:\n")

			for _, call in ipairs(hlog.utils.sorted_keys(cur.p2p)) do
				local refs = cur.p2p[call]
				table.sort(refs)

				local out = string.format("  %s (%d) @",
							  call, hlog.utils.count_items(refs))
				for _, ref in ipairs(refs) do
					out = out .. " " .. ref
				end
				f:write(out .. "\n")
			end
			f:write("\n")
		end

		-- dxcc stats
		write_dxcc_stats(f, cur.dxcc)

		-- state stats
		write_state_stats(f, cur.states)

		-- ADIF header
		f:write(hlog.adif.write_header())
		f:write("\n")

		-- ADIF QSOs
		for _, qso in pairs(cur.chrono) do
			-- note the QSO uuid
			uuids = open_uuid_file(uuids, logpath, ".uuids")

			uuids:write(qso.uuid .. "\n")

			-- Do *not* use strict mode, POTA doesn't check the submode field
			-- and the strict mode field doesn't include enough info (e.g.,
			-- PSK31 and QPSK31 both use "PSK" as the mode).
			f:write(hlog.adif.write_qso(qso, false))
			f:write("\n")
		end

		f:close()
		close_uuid_file(uuids)
	end
end

local function export(outdir)
	local pota = {
		logs = {},
		mangled_qso_to_orig = {},
	}
	local wwff = {
		logs = {},
		mangled_qso_to_orig = {},
	}
	local sota = {
		logs = {},
		mangled_qso_to_orig = {},
	}

	for qso in hlog.index.history(true) do
		add_entry(qso, pota, 'POTA', 'tx_pota', 'rx_pota', 'pota_log_submitted', true)
		add_entry(qso, wwff, 'WWFF', 'tx_wwff', 'rx_wwff', 'wwff_log_submitted', false)
		add_entry(qso, sota, 'SOTA', 'tx_sota', 'rx_sota', 'sota_log_submitted', false)
	end

	gen_logs(pota, 'POTA', outdir)
	gen_logs(wwff, 'WWFF', outdir)
	gen_logs(sota, 'SOTA', outdir)
end

return {
	labels = {
		{ 0,  0, "St. Call" },
		{ 0, 11, "RSTs" },
		{ 0, 16, "RSTr" },
		{ 0, 21, "St" },
		{ 0, 24, "Name" },
		{ 0, 41, "Refs" },
	},

	fields = {
		{ 1,  0, 10, "rx.station_call", },
		{ 1, 11,  4, "tx.rst", },
		{ 1, 16,  4, "rx.rst", },
		{ 1, 21,  2, "rx.state", },
		{ 1, 24, 16, "name", },
		{ 1, 41, 15, "ref", },
	},

	annunciators = {
		{ 0, 57, "DUP", "black" },
		{ 0, 61, "AM", "black" }, -- early shift
		{ 1, 61, "PM", "black" }, -- late shift
	},

	events = {
		-- called once when exporting the contest log
		export = export,

		-- called once on startup (after config's fill_template)
		startup = function(template)
			print("POTA contest script")
			startup_fill_template(template)
		end,

		-- called when QSO is being initialized
		--
		-- We rely on the tx_{pota,wwff} additional fields already
		-- being set by the template.
		qso_init = function(qso)
			local ref = qso.additional['tx_pota']

			contest.lcd(ref ~= nil and ref or "?")
		end,

		-- called when QSO is complete & saved, just prior to a reset
		qso_done = qso_done,

		-- called whenever a contest field changes
		field_changed = field_changed,
	},
}
