/*
 * Copyright (c) 2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "xlua_impl.h"

/* add in scriptname key-values to table at the top of the stack */
static void merge_lua_funcs(lua_State *L, const char *scriptname)
{
	const int orig_top = lua_gettop(L);
	char error[128];
	int ret;

	ASSERT3U(lua_type(L, -1), ==, LUA_TTABLE);

	ret = xlua_load_script_internal(L, scriptname, error, sizeof(error));
	if (ret)
		panic("Failed to load hlog.utils script fragment '%s': %s\n",
		      scriptname, error);

	switch (lua_type(L, -1)) {
		case LUA_TTABLE:
			/* ok, merge it */
			lua_pushnil(L);
			while (lua_next(L, -2)) {
				/* stack: utils table, this table, key, value */

				if (!lua_isstring(L, -2) ||
				    !lua_isfunction(L, -1))
					panic("Warning: hlog.utils fragment "
					      "'%s' table contains %s=%s pair\n",
					      scriptname,
					      lua_typename(L, lua_type(L, -2)),
					      lua_typename(L, lua_type(L, -1)));

				/* duplicate the key */
				lua_pushvalue(L, -2);
				lua_insert(L, -2);

				/* stack: utils table, table, key, key, value */

				/* set the k=v on the utils table */
				lua_settable(L, -5);
			}

			/* pop the table */
			lua_pop(L, 1);
			break;
		case LUA_TNIL:
			/* nothing, weird but ok */
			lua_pop(L, 1);
			break;
		default:
			/* other types */
			panic("Unknown type %s gotten from hlog.utils fragment "
			      "'%s'\n", lua_typename(L, lua_type(L, -1)),
			      scriptname);
	}

	ASSERT_LUA_STACK(L, orig_top);
}

static const luaL_Reg utils_funcs[] = {
	{ NULL, NULL },
};

void xlua_pushlib_utils(lua_State *L)
{
	/* set functions implemented in C */
	luaL_newlib(L, utils_funcs);

	/* set functions implemented in lua */
	merge_lua_funcs(L, "utils-math.lua");
	merge_lua_funcs(L, "utils-split.lua");
	merge_lua_funcs(L, "utils-table.lua");
}
